﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace AceGrep
{
    public class FileNumberLineLimitedTable : Perusable
    {
        int column, maximumLineCount, nameLength;
        string wideSpace;

        private int fixedWidthFontScreenColumn(string name)
        {
            int q;
            int size = 0;
            string c;

            for (int i = 0; i < name.Length; ++i)
            {
                c = name.Substring(i, 1);
                q = ASCII_KATAKANA.IndexOf(c);
                if (q >= 0)
                {
                    size += ASCII_SIZE;
                }
                else
                {
                    size += KANJI_SIZE;
                }

            }

            return size;
        }

        private string separating(string name, int lineNumber, string line)
        {
            string line_number = lineNumber.ToString();
            int line_number_length = line_number.Length;
            int q = column - line_number_length;
            int r = column - line_number_length - name.Length;
            int subtotal = nameLength + GAP_SIZE + line_number_length;
            string row;

            if (subtotal < column)
            {
                row = name + wideSpace.Substring(nameLength, r) + line_number + separator + line;
            }
            else
            {
                int abandoned;

                if (line_number_length < column)
                {
    // Grep abandons a part of the name so that the name-gap length can equal q.
                    string nameGap = name + GAP;
                    abandoned = nameGap.Length - q;
                    row = nameGap.Substring(abandoned) + line_number + separator + line;
                }
                else
                {
    // Grep abandons a part of the line number so that the line number length can equal column.
                    abandoned = line_number_length - column;
                    row = line_number.Substring(abandoned) + separator + line;
                }

            }

            return row;
        }

        public override bool reading()
        {
            bool found = false;
            int count = 0;
            string line = EMPTY;
            string name = label;
            nameLength = fixedWidthFontScreenColumn(name);

            try
            {
                for (int lineNumber = 1; count < maximumLineCount; lineNumber += 1)
                {
                    line = Console.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    m = x.Match(line);
                    if (inverted ^ m.Success)
                    {
                        Console.WriteLine( separating(name, lineNumber, line) );
                        count += 1;
                    }

                }

                if(count > 0) {
                    found = true;
                }

            }
            catch (IOException e133)
            {
                Error.handler(e133.ToString());
                plus("IOException e133");
            }

            return found;
        }

        public override bool reading(string name)
        {
            bool found = false;
            int count = 0;
            string line = EMPTY;
            nameLength = fixedWidthFontScreenColumn(name);

            try
            {
                StreamReader file = new StreamReader(name, Encoding.Default);

                for (int lineNumber = 1; count < maximumLineCount; lineNumber += 1)
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    m = x.Match(line);
                    if (inverted ^ m.Success)
                    {
                        Console.WriteLine(separating(name, lineNumber, line));
                        count += 1;
                    }

                }

                file.Close();
                if(count > 0) {
                    found = true;
                }

            }
            catch (IOException e134)
            {
                Error.handler(e134.ToString());
                plus("IOException e134");
            }

            return found;
        }

        public override bool reading(string name, Encoding encoding)
        {
            bool found = false;
            int count = 0;
            string line = EMPTY;
            nameLength = fixedWidthFontScreenColumn(name);

            try
            {
                StreamReader file = new StreamReader(name, encoding);

                for (int lineNumber = 1; count < maximumLineCount; lineNumber += 1)
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    m = x.Match(line);
                    if (inverted ^ m.Success)
                    {
                        Console.WriteLine(separating(name, lineNumber, line));
                        count += 1;
                    }

                }

                file.Close();
                if(count > 0) {
                    found = true;
                }

            }
            catch (IOException e135)
            {
                Error.handler(e135.ToString());
                plus("IOException e135");
            }

            return found;
        }

        public FileNumberLineLimitedTable(string re, int c, int m, bool insensitive)
        {
            case_sensitive = !insensitive;

            column = c;
            maximumLineCount = m;
            wideSpace = EMPTY;
            for (int i = 0; i < column; ++i)
            {
                wideSpace += GAP;
            }

            if (compiling(re))
            {
                error = EMPTY;
            }
            else
            {
                error = "compiling(re): /" + re + "/";
            }

        }

    }
}
