﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Security;

namespace AceGrep
{
    class Grep
    {
        const bool NO_APPENDING = false;
        const bool NEITHER_CARET_NOR_DOLLAR = false;
        const bool SERIAL_NUMBER = true;
        const double VERSION = 1.01;
        const int ASCII_GREP = 3;
        const int BINARY_GREP = 0;
        const int CHAR_ARRAY_SIZE = 4096;
        const int CP_SIZE = 2;
        const int DIR = 1;
        const int ENCODED_GREP = 1;
        const int ENCODING_EQUALS_SIZE = 11;
        const int ENCODING_SIZE = 1;
        const int ERROR = 2;
        const int FIELD_WIDTH = 48;
        const int FOUND = 0;
        const int INTERNAL = 0; // Grep selects DIR or ls if command exceeds zero.
        const int IO_ERROR = 5;
        const int LABEL_EQUALS_SIZE = 8;
        const int LS = 2;
        const int MANY_TIMES = 40; // -b40
        const int MAXIMUM_LINE_COUNT = 0x7FFFFFFF;
        const int MIXED_GREP = 2;
        const int NOT_CONVERTED = -1;
        const int OTHERWISE = 1;
        const int QUEUE_ARRAY_SIZE = 384;
        const int REGEX_ERROR = 3;
        const int RING_COUNT = QUEUE_ARRAY_SIZE - 1;
        const int SAFE_AND_SOUND = 0;
        const int SECURITY_ERROR = 6;
        const int SLOW = 600;
        const int WAITING_FOR_A_FILE = 100;
        const int WAITING_UNTIL_FILE_IS_DELETED = 100;
        const long TOO_SMALL = 1L;
        const string ACE = "--ace";
        const string ADDITIONAL_REGULAR_EXPRESSION_SEQUENCE_END = "-e-";
        const string ADDITIONAL_REGULAR_EXPRESSION_SEQUENCE_START = "-e";
        const string AFTER = "-A";
        const string ASCII_DASH = "--ascii";
        const string ASCII_YES = "-y";
        const string BATCH = "-b";
        const string BEFORE = "-B";
        const string C_CHCP = "/C CHCP";
        const string C_DIR_B = "/C DIR /B ";
        const string C_DIR_B_BATCH = "/C C:\\AceGrep\\DIR_B.bat";
        const string C_DIR_S_B = "/C DIR /S /B ";
        const string C_DIR_S_B_BATCH = "/C C:\\AceGrep\\DIR_S_B.bat";
        const string C_LS_1 = "/C ls -1 ";
        const string C_LS_1_BATCH = "/C C:\\AceGrep\\LS_1.bat";
        const string C_LS_R_1 = "/C ls -R -1 ";
        const string C_LS_R_1_BATCH = "/C C:\\AceGrep\\LS_R_1.bat";
        const string CARET = "^";
        const string CARET_PARENTHESIS = "^(";
        const string CHCP_TEXT = "C:\\AceGrep\\CHCP.txt";
        const string CLOSE_DOLLAR = ")$";
        const string CMD_EXE = "cmd.exe";
        const string CODE_PAGE = "-#";
        const string COLON_GAP = ": ";
        const string CONTEXT = "-C";
        const string CONTROL_CHARACTERS = "-[";
        const string COUNT = "-c";
        const string COUNT_EQUALS_ZERO = "-L";
        const string COUNT_EXCEEDS_ZERO = "--ce";
        const string CP = "CP";
        const string CURRENT_DIRECTORY = "-g";
        const string DIR_B = "DIR /B ";
        const string DIR_B_BAT = "C:\\AceGrep\\DIR_B.bat";
        const string DIR_S_B = "DIR /S /B ";
        const string DIR_S_B_BAT = "C:\\AceGrep\\DIR_S_B.bat";
        const string DIR_TERM = "--dir";
        const string DOLLAR = "$";
        const string DOUBLE_SPACED = "\x0D\x0A\x0D\x0A";
        const string ELAPSE = "--elapse";
        const string EMPTY = "";
        const string ENCODING = "<";
        const string ENCODING_EQUALS = "--encoding=";
        const string ERROR_TEXT = "error.txt";
        const string EUC_JP = "EUC-JP";
        const string EXTERNAL_BATCH = "--bat";
        const string FILE_NAME_FOR_EACH_MATCH = "-H";
        const string FILE_NAME_LIST = "FNL";
        const string FILE_NAME_LIST_TEXT = "fileName.txt";
        const string FILE_NAMES_REQUIRED = "-l";
        const string GAP = " ";
        const string HELP = "--help";
        const string HYPHEN = "-";
        const string I_O_EXCEPTION = "IOException";
        const string INSENSITIVE = "-i";
        const string INTERNAL_TERM = "--in";
        const string INVERTED = "-v";
        const string JIKOKU = "-j"; // time
        const string KANJI_8 = "-8";
        const string KANJI_EUC_JP = "-Ke";
        const string KANJI_SHIFT_JIS = "-Ks";
        const string KANJI_UTF_8 = "-Ku";
        const string KANJI_WINDOWS31J = "-Kw";
        const string KUDA = "-k"; // pipe
        const string LABEL_EQUALS = "--label=";
        const string LINE_NUMBER = "-n";
        const string LS_1 = "ls -1 ";
        const string LS_1_BAT = "C:\\AceGrep\\LS_1.bat";
        const string LS_R_1 = "ls -R -1 ";
        const string LS_R_1_BAT = "C:\\AceGrep\\LS_R_1.bat";
        const string MAXIMUM = "-m";
        const string MILLISECONDS = "-@";
        const string MS932 = "MS932";
        const string N_0TO9 = "0123456789";
        const string NEW_LINE = "\x0D\x0A";
        const string NO_BATCH = "-b-";
        const string NO_ERROR = "no error";
        const string NO_LOGO = "";
        const string NO_PATH = "-p-";
        const string OBTAIN_PATTERNS_FROM_FILE = "-f";
        const string ONE_MILLISECOND = "1";
        const string PATTERNS_TEXT = "patterns.txt";
        const string QUIET = "-q";
        const string QUOTE = "'"; // single quote
        const string RECURSIVELY_LIST_ALL_FILES = "-r";
        const string REDIRECT = " > ";
        const string REGULAR_EXPRESSION = "[01234][0-9] ";
        const string SEPARATE_TEXT = "separate.txt";
        const string SEPARATOR = "-S";
        const string SHIFT_JIS = "Shift_JIS";
        const string SHOW_REGULAR_EXPRESSION = "--re";
        const string SIDE_ACE_EXE = "SideAce.exe";
        const string SPECIAL_FILE_NAME = "-";
        const string STANDARD_INPUT = "stdin";
        const string TABLE = "-T";
        const string TEXT_EDITOR = "-I";
        const string TEXT_FILE = ".txt";
        const string TICK = "--tick";
        const string UNIX = "-u";
        const string UTF_8 = "UTF-8";
        const string VERSION_PHRASE = "--version";
        const string VERSION_TERM = "-V";
        const string WHOLE_LINE = "-x";
        const string YEN_QUOTE = "\\'"; // \'

        bool anotherRegularExpression, countEqualsZero, countExceedsZero, elapsed, externalBatch, fileNameForEachMatch, fileNamesRequired, found, insensitive, inverted, lineNumberFollowed, listOfPatterns, matchCount, newListShallBeProvided, noListing, reduceToCurrentDirectory, showRegularExpression, specialFileName, standardInput, textEditor, throughSubdirectories, ticktock, wholeLine;
        bool dir = true; // command = DIR
        bool examples = true;
        bool fullPath = true;
        bool loud = true;
        bool replaceControlCharacters = true;
        bool windows = true;
        char[] char_array; // TextReader.Read(), TextReader.Write()
        Dictionary<string, Encoding> d = new Dictionary<string, Encoding>();
        int after = 0;
        int before = 0;
        int command = DIR; // dir = true
        int fieldWidth = 0;
        int manager = MIXED_GREP;
        int manyTimes = MANY_TIMES;
        int maximumLineCount = MAXIMUM_LINE_COUNT;
        int regularExpressionCount = 0;
        int ringCount = 1;
        string argument = EMPTY;
        string fileNameListText = FILE_NAME_LIST_TEXT;
        string inputEncoding = MS932;
        string label = STANDARD_INPUT;
        string patternSyntax = NO_ERROR;
        string re = REGULAR_EXPRESSION;
        string regularExpressionText = PATTERNS_TEXT;
        string separator = COLON_GAP;
        string wildCards = "*.*";

        private bool deleteFileNameText()
        {
            bool b = true;

            try
            {
                File.Delete(FILE_NAME_LIST_TEXT);
                Thread.Sleep(WAITING_UNTIL_FILE_IS_DELETED);
            }
            catch (IOException e9)
            {
                b = false;
                Error.handler(e9.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e9");
                }

            }

            return b;
        }

        private bool inquire(string encoding_method)
        {
            bool b = true;

            try
            {
                Encoding encoding = d[encoding_method];
            }
            catch (KeyNotFoundException e178)
            {
                b = false;
                Error.handler(e178.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("KeyNotFoundException e178");
                    System.Environment.Exit(ERROR);
                }
                else
                {
                    System.Environment.Exit(OTHERWISE);
                }

            }

            return b;
        }

        private bool makeBatchFile(bool subdirectories, string f, string fileNameListText)
        {
            bool b = true;
            StreamWriter file;

            try
            {
                if (subdirectories)
                {
                    if (windows)
                    {
                        file = new StreamWriter(DIR_S_B_BAT, NO_APPENDING, Encoding.Default);
                        file.Write(DIR_S_B);
                        file.Write(f);
                    }
                    else
                    {
                        // Typically, adapting() can adapt the redirected text (fileNameListText) to Grep.
                        file = new StreamWriter(LS_R_1_BAT, NO_APPENDING, Encoding.Default);
                        file.Write(LS_R_1);
                    }

                    file.Write(REDIRECT);
                    file.Write(fileNameListText);
                    file.Close();
                }
                else
                {
                    if (windows)
                    {
                        file = new StreamWriter(DIR_B_BAT, NO_APPENDING, Encoding.Default);
                        file.Write(DIR_B);
                    }
                    else
                    {
                        file = new StreamWriter(LS_1_BAT, NO_APPENDING, Encoding.Default);
                        file.Write(LS_1);
                    }

                    file.Write(f);
                    file.Write(REDIRECT);
                    file.Write(fileNameListText);
                    file.Close();
                }

            }
            catch (IOException e10)
            {
                b = false;
                Error.handler(e10.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e10");
                }

            }

            return b;
        }

        private bool preparingSeparateText()
        {
            bool b = true;

            try
            {
                // SEPARATE_TEXT
                StreamWriter separateFile = new StreamWriter(SEPARATE_TEXT, NO_APPENDING, Encoding.Default);
                separateFile.Write(COLON_GAP);
                separateFile.Close();
            }
            catch (IOException e11)
            {
                b = false;
                Error.handler(e11.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e11");
                }

            }

            return b;
        }

        private bool sideAce()
        {
            bool b = true;

            try
            {
                Process p = new Process();
                p.StartInfo.FileName = SIDE_ACE_EXE;
                p.StartInfo.Arguments = EMPTY;
                p.Start();
            }
            catch (SecurityException e12)
            {
                b = false;
                if (loud)
                {
                    Console.Error.WriteLine(e12.ToString());
                }

            }
            catch (IOException e13)
            {
                b = false;
                if (loud)
                {
                    Console.Error.WriteLine(e13.ToString());
                }

            }

            return b;
        }

        private bool CHCP()
        {
            bool b = true;
            int read_size;

            try
            {
                // cmd.exe /C CHCP > CHCP.txt
                Process p = new Process();
                p.StartInfo.FileName = CMD_EXE;
                p.StartInfo.Arguments = C_CHCP;
                p.StartInfo.UseShellExecute = false;
                p.StartInfo.RedirectStandardOutput = true;
                p.Start();
                StreamReader redirected = p.StandardOutput;
                StreamWriter chcpFile = new StreamWriter(CHCP_TEXT, NO_APPENDING, Encoding.Default);

                while (!redirected.EndOfStream)
                {
                    char_array = new char[CHAR_ARRAY_SIZE];
                    read_size = redirected.Read(char_array, 0, CHAR_ARRAY_SIZE);
                    chcpFile.Write(char_array, 0, read_size);
                }

                p.Close();

                chcpFile.Close();
            }
            catch (SecurityException e14)
            {
                b = false;
                if (loud)
                {
                    Console.Error.WriteLine(e14.ToString());
                }

            }
            catch (IOException e15)
            {
                b = false;
                if (loud)
                {
                    Console.Error.WriteLine(e15.ToString());
                }

            }

            return b;
        }

        private bool checkingCHCP_TEXT()
        {
            bool b = true;

            try
            {
                int q;
                StreamReader chcpFile = new StreamReader(CHCP_TEXT, Encoding.Default);
                string c;
                string s = chcpFile.ReadLine();

                for (int i = 0; i < s.Length; ++i)
                {
                    c = s.Substring(i, 1);
                    q = N_0TO9.IndexOf(c);
                    if (q >= 0)
                    {
                        Console.Write(c);
                    }

                }

                Console.WriteLine();
                chcpFile.Close();
            }
            catch (IOException e16)
            {
                b = false;
                Error.handler(e16.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e16");
                }

            }

            return b;
        }

        private bool tryToOpenHyphenFile(string s)
        {
            bool b = true;

            try
            {
                StreamReader hyphenFile = new StreamReader(s, Encoding.Default);
                hyphenFile.Close();
            }
            catch (IOException e19)
            {
                b = false;
                Error.handler(e19.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e19");
                }

            }

            return b;
        }

        private int asciiGrep(Perusable r)
        {
            int exitCode = FOUND;
            string fileName;

            try
            {
                StreamReader listFile = new StreamReader(fileNameListText, Encoding.Default);
                for (; ; )
                {
                    fileName = listFile.ReadLine();
                    if (fileName == null)
                    {
                        break;
                    }

                    if (FileAttributes.Directory == File.GetAttributes(fileName))
                    {
                        continue;
                    }

                    found = r.reading(fileName);
                }

                listFile.Close();
            }
            catch (IOException e20)
            {
                Error.handler(e20.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e20");
                    exitCode = ERROR;
                }
                else
                {
                    if (found)
                    {
                        exitCode = FOUND;
                    }
                    else
                    {
                        exitCode = OTHERWISE;
                    }

                }

            }

            return exitCode;
        }

        private int binaryGrep(Perusable r)
        {
            Encoding encoding = d[inputEncoding];
            int exitCode = FOUND; ;
            int q;
            string fileName, input_encoding, s;

            try
            {
                StreamReader listFile = new StreamReader(fileNameListText, Encoding.Default);
                for (; ; )
                {
                    s = listFile.ReadLine();
                    if (s == null)
                    {
                        break;
                    }

                    q = s.IndexOf(ENCODING);
                    if (q >= 0)
                    {
                        fileName = s.Substring(0, q);
                        q += ENCODING_SIZE;
                        input_encoding = s.Substring(q);
                        encoding = d[input_encoding];

                        if (FileAttributes.Directory == File.GetAttributes(fileName))
                        {
                            continue;
                        }

                        if (textEditor)
                        {
                            if (r.tooManyControlCharacters(fileName))
                            {
                                continue;
                            }

                        }

                        if (replaceControlCharacters)
                        {
                            fileName = r.replaceControlCharactersWithCaretCapitals(fileName);
                            if (I_O_EXCEPTION.CompareTo(fileName) == 0)
                            {
                                continue;
                            }

                        }

                        found = r.reading(fileName, encoding);
                    }
                    else
                    {
                        fileName = s;

                        if (FileAttributes.Directory == File.GetAttributes(fileName))
                        {
                            continue;
                        }

                        if (textEditor)
                        {
                            if (r.tooManyControlCharacters(fileName))
                            {
                                continue;
                            }

                        }

                        if (replaceControlCharacters)
                        {
                            fileName = r.replaceControlCharactersWithCaretCapitals(fileName);
                            if (I_O_EXCEPTION.CompareTo(fileName) == 0)
                            {
                                continue;
                            }

                        }

                        found = r.reading(fileName);
                    }

                }

                listFile.Close();
            }
            catch (IOException e21)
            {
                Error.handler(e21.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e21");
                    exitCode = ERROR;
                }
                else
                {
                    if (found)
                    {
                        exitCode = FOUND;
                    }
                    else
                    {
                        exitCode = OTHERWISE;
                    }

                }

            }
            catch (KeyNotFoundException e22)
            {
                Error.handler(e22.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("KeyNotFoundException e22");
                    exitCode = ERROR;
                }
                else
                {
                    if (found)
                    {
                        exitCode = FOUND;
                    }
                    else
                    {
                        exitCode = OTHERWISE;
                    }

                }

            }

            return exitCode;
        }

        private int encodedGrep(Perusable r)
        {
            Encoding encoding = d[inputEncoding];
            int exitCode = FOUND;
            int q;
            string fileName, s;

            try
            {
                StreamReader listFile = new StreamReader(fileNameListText, Encoding.Default);
                for (; ; )
                {
                    s = listFile.ReadLine();
                    if (s == null)
                    {
                        break;
                    }

                    q = s.IndexOf(ENCODING);
                    if (q >= 0)
                    {
                        fileName = s.Substring(0, q);
                    }
                    else
                    {
                        fileName = s;
                    }

                    if (FileAttributes.Directory == File.GetAttributes(fileName))
                    {
                        continue;
                    }

                    found = r.reading(fileName, encoding);
                }

                listFile.Close();
            }
            catch (IOException e23)
            {
                Error.handler(e23.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e23");
                    exitCode = ERROR;
                }
                else
                {
                    if (found)
                    {
                        exitCode = FOUND;
                    }
                    else
                    {
                        exitCode = OTHERWISE;
                    }

                }

            }
            catch (KeyNotFoundException e24)
            {
                Error.handler(e24.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("KeyNotFoundException e24");
                    exitCode = ERROR;
                }
                else
                {
                    if (found)
                    {
                        exitCode = FOUND;
                    }
                    else
                    {
                        exitCode = OTHERWISE;
                    }

                }

            }

            return exitCode;
        }

        private int mixedGrep(Perusable r)
        {
            Encoding encoding = d[inputEncoding];
            int exitCode = FOUND;
            int q;
            string fileName, input_encoding, s;

            try
            {
                StreamReader listFile = new StreamReader(fileNameListText, Encoding.Default);
                for (; ; )
                {
                    s = listFile.ReadLine();
                    if (s == null)
                    {
                        break;
                    }

                    q = s.IndexOf(ENCODING);
                    if (q >= 0)
                    {
                        fileName = s.Substring(0, q);
                        q += ENCODING_SIZE;
                        input_encoding = s.Substring(q);
                        encoding = d[input_encoding];

                        if (FileAttributes.Directory == File.GetAttributes(fileName))
                        {
                            continue;
                        }

                        found = r.reading(fileName, encoding);
                    }
                    else
                    {
                        fileName = s;

                        if (FileAttributes.Directory == File.GetAttributes(fileName))
                        {
                            continue;
                        }

                        found = r.reading(fileName);
                    }

                }

                listFile.Close();
            }
            catch (IOException e25)
            {
                Error.handler(e25.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e25");
                    exitCode = ERROR;
                }
                else
                {
                    if (found)
                    {
                        exitCode = FOUND;
                    }
                    else
                    {
                        exitCode = OTHERWISE;
                    }

                }

            }
            catch (KeyNotFoundException e26)
            {
                Error.handler(e26.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("KeyNotFoundException e26");
                    exitCode = ERROR;
                }
                else
                {
                    if (found)
                    {
                        exitCode = FOUND;
                    }
                    else
                    {
                        exitCode = OTHERWISE;
                    }

                }

            }

            return exitCode;
        }

        private int provideFileNameListText(bool subdirectories)
        {
            AceList a;
            int exitCode = SAFE_AND_SOUND;
            int max = 0;
            string[] items = wildCards.Split(' ');

            switch (command)
            {
                case DIR:
                    exitCode = useDIR();
                    break;
                case LS:
                    exitCode = use_ls();
                    break;
                default:
                    File.Delete(fileNameListText);
                    foreach (string s in items)
                    {
                        if ( EMPTY.CompareTo(s) == 0 )
                        {
                            continue;
                        }

                        a = new AceList(fileNameListText, s, reduceToCurrentDirectory, showRegularExpression, subdirectories);
                        exitCode = a.internalMethod(s);
                        if (exitCode > max)
                        {
                            max = exitCode;
                        }

                    }

                    break;
            }

            return exitCode;
        }

        private int selected(bool dir, bool windows)
        {
            int k = DIR;

            if (windows)
            {
                if (dir)
                {
                    k = DIR;
                    if (throughSubdirectories)
                    {
                        if (reduceToCurrentDirectory)
                        {
                            throughSubdirectories = false;
                        }

                    }
                    else
                    {
                        reduceToCurrentDirectory = true;
                    }

                }
                else
                {
                    k = INTERNAL;
                    if (throughSubdirectories)
                    {
                        if (reduceToCurrentDirectory)
                        {
                            throughSubdirectories = false;
                        }

                    }
                    else
                    {
                        reduceToCurrentDirectory = true;
                    }

                }

            }
            else
            {
                k = LS;
            }

            return k;
        }

        private int sequence_to_integer(string s)
        {
            int a = NOT_CONVERTED;
            try
            {
                a = Convert.ToInt32(s);
            }
            catch (FormatException e179)
            {
                // ERROR_TEXT
                StreamWriter errorFile = new StreamWriter(ERROR_TEXT, NO_APPENDING, Encoding.Default);
                errorFile.WriteLine(e179.ToString());
                errorFile.Close();
            }

            return a;
        }

        public int run()
        {
            Elapse e = new Elapse();
            int exitCode = OTHERWISE;
            int patternCount = 0;
            Perusable r = new MatchCount(re, insensitive);
            Stopwatch s = Stopwatch.StartNew();
            Stopwatch w = Stopwatch.StartNew();
            string milliseconds = EMPTY;
            StringBuilder or = new StringBuilder();
            int q = inputEncoding.IndexOf(CP);
            if (q >= 0)
            {
                addCodePage(inputEncoding, q);
            }

            d.Add("ASCII", Encoding.ASCII);
            d.Add("EUC_JP", Encoding.GetEncoding("euc-jp"));
            d.Add("ISO-8859-1", Encoding.GetEncoding(850));
            d.Add("ISO8859_1", Encoding.GetEncoding(850));
            d.Add("JIS", Encoding.GetEncoding("iso-2022-jp"));
            d.Add("SJIS", Encoding.GetEncoding("shift_jis"));
            d.Add("Unicode", Encoding.Unicode);
            d.Add("US_ASCII", Encoding.GetEncoding("us-ascii"));
            d.Add("UTF-16", Encoding.Unicode);
            d.Add("UTF-32", Encoding.UTF32);
            d.Add("UTF-7", Encoding.UTF7);
            d.Add("UTF8", Encoding.UTF8);
            d.Add(EUC_JP, Encoding.GetEncoding("euc-jp"));
            d.Add(MS932, Encoding.Default);
            d.Add(SHIFT_JIS, Encoding.GetEncoding("shift_jis"));
            d.Add(UTF_8, Encoding.UTF8);
            inquire(inputEncoding);

            if (ticktock)
            {
                s = Stopwatch.StartNew();
            }

            if (elapsed)
            {
                e.setting();
            }

            if (listOfPatterns)
            {
                try
                {
                    StreamReader patternFile = new StreamReader(regularExpressionText, Encoding.Default);
                    for (; ; )
                    {
                        re = patternFile.ReadLine();
                        if (re == null)
                        {
                            break;
                        }

                        patternCount += 1;
                    }

                    patternFile.Close();
                }
                catch (FileNotFoundException e27)
                {
                    if (loud)
                    {
                        Console.Error.WriteLine("FileNotFoundException e27");
                        Console.Error.WriteLine(e27.ToString());
                        System.Environment.Exit(ERROR);
                    }
                    else
                    {
                        System.Environment.Exit(OTHERWISE);
                    }

                }

                string[] pattern = new string[patternCount];
                try
                {
                    StreamReader patternFile = new StreamReader(regularExpressionText, Encoding.Default);
                    for (int i = 0; i < patternCount; ++i)
                    {
                        re = patternFile.ReadLine();
                        if (re == null)
                        {
                            break;
                        }

                        pattern[i] = re;
                    }

                    patternFile.Close();
                }
                catch (IOException e28)
                {
                    if (loud)
                    {
                        Console.Error.WriteLine("IOException e28");
                        Console.Error.WriteLine(e28.ToString());
                        System.Environment.Exit(ERROR);
                    }
                    else
                    {
                        System.Environment.Exit(OTHERWISE);
                    }

                }

                if (fileNamesRequired)
                {
                    if (matchCount)
                    {
                        if (countExceedsZero)
                        {
                            r = new MultiPatternMatchCountExceedsZero(pattern, insensitive); // String[] pattern
                        }
                        else
                        {
                            if (countEqualsZero)
                            {
                                r = new MultiPatternMatchCountEqualsZero(pattern, insensitive);
                            }
                            else
                            {
                                r = new MultiPatternMatchCount(pattern, insensitive);
                            }

                        }

                    }
                    else
                    {
                        r = new MultiPatternFileRequired(pattern, insensitive);
                    }

                }
                else
                {
                    if (maximumLineCount == MAXIMUM_LINE_COUNT)
                    {
                        if (fileNameForEachMatch)
                        {
                            if (lineNumberFollowed)
                            {
                                if (fieldWidth > 0)
                                {
                                    r = new MultiPatternFileNumberLineTable(pattern, fieldWidth, insensitive);
                                }
                                else
                                {
                                    r = new MultiPatternFileNumberLine(pattern, insensitive);
                                }

                            }
                            else
                            {
                                r = new MultiPatternFileLine(pattern, insensitive);
                            }

                        }
                        else
                        {
                            if (lineNumberFollowed)
                            {
                                if (fieldWidth > 0)
                                {
                                    r = new MultiPatternNumberLineTable(pattern, fieldWidth, insensitive);
                                }
                                else
                                {
                                    r = new MultiPatternNumberLine(pattern, insensitive);
                                }

                            }
                            else
                            {
                                if (ringCount > 1)
                                {
                                    r = new MultiPatternLineRing(pattern, after, before, insensitive);
                                }
                                else
                                {
                                    r = new MultiPatternLine(pattern, insensitive);
                                }

                            }

                        }

                    } else {
                        if (fileNameForEachMatch)
                        {
                            if (lineNumberFollowed)
                            {
                                if (fieldWidth > 0)
                                {
                                    r = new MultiPatternFileNumberLineLimitedTable(pattern, fieldWidth, maximumLineCount, insensitive);
                                }
                                else
                                {
                                    r = new MultiPatternFileNumberLineLimited(pattern, maximumLineCount, insensitive);
                                }

                            }
                            else
                            {
                                r = new MultiPatternFileLineLimited(pattern, maximumLineCount, insensitive);
                            }

                        }
                        else
                        {
                            if (lineNumberFollowed)
                            {
                                if (fieldWidth > 0)
                                {
                                    r = new MultiPatternNumberLineLimitedTable(pattern, fieldWidth, maximumLineCount, insensitive);
                                }
                                else
                                {
                                    r = new MultiPatternNumberLineLimited(pattern, maximumLineCount, insensitive);
                                }

                            }
                            else
                            {
                                if (ringCount > 1)
                                {
                                    r = new MultiPatternLineLimitedRing(pattern, after, before, maximumLineCount, insensitive);
                                }
                                else
                                {
                                    r = new MultiPatternLineLimited(pattern, maximumLineCount, insensitive);
                                }

                            }

                        }

                    }

                }

            }
            else
            {
                if (fileNamesRequired)
                {
                    if (matchCount)
                    {
                        if (countExceedsZero)
                        {
                            r = new MatchCountExceedsZero(re, insensitive);
                        }
                        else
                        {
                            if (countEqualsZero)
                            {
                                r = new MatchCountEqualsZero(re, insensitive);
                            }
                            else
                            {
                                r = new MatchCount(re, insensitive);
                            }

                        }

                    }
                    else
                    {
                        r = new FileRequired(re, insensitive);
                    }

                }
                else
                {
                    if (maximumLineCount == MAXIMUM_LINE_COUNT)
                    {
                        if (fileNameForEachMatch)
                        {
                            if (lineNumberFollowed)
                            {
                                if (fieldWidth > 0)
                                {
                                    r = new FileNumberLineTable(re, fieldWidth, insensitive);
                                }
                                else
                                {
                                    r = new FileNumberLine(re, insensitive);
                                }

                            }
                            else
                            {
                                r = new FileLine(re, insensitive);
                            }

                        }
                        else
                        {
                            if (lineNumberFollowed)
                            {
                                if (fieldWidth > 0)
                                {
                                    r = new NumberLineTable(re, fieldWidth, insensitive);
                                }
                                else
                                {
                                    r = new NumberLine(re, insensitive);
                                }

                            }
                            else
                            {
                                if (ringCount > 1)
                                {
                                    r = new LineRing(re, after, before, insensitive);
                                }
                                else
                                {
                                    r = new Line(re, insensitive);
                                }

                            }

                        }

                    }
                    else
                    {
                        if (fileNameForEachMatch)
                        {
                            if (lineNumberFollowed)
                            {
                                if (fieldWidth > 0)
                                {
                                    r = new FileNumberLineLimitedTable(re, fieldWidth, maximumLineCount, insensitive);
                                }
                                else
                                {
                                    r = new FileNumberLineLimited(re, maximumLineCount, insensitive);
                                }

                            }
                            else
                            {
                                r = new FileLineLimited(re, maximumLineCount, insensitive);
                            }

                        }
                        else
                        {
                            if (lineNumberFollowed)
                            {
                                if (fieldWidth > 0)
                                {
                                    r = new NumberLineLimitedTable(re, fieldWidth, maximumLineCount, insensitive);
                                }
                                else
                                {
                                    r = new NumberLineLimited(re, maximumLineCount, insensitive);
                                }

                            }
                            else
                            {
                                if (ringCount > 1)
                                {
                                    r = new LineLimitedRing(re, after, before, maximumLineCount, insensitive);
                                }
                                else
                                {
                                    r = new LineLimited(re, maximumLineCount, insensitive);
                                }

                            }

                        }

                    }

                }

            }

            r.setting(label, separator, inverted);
            patternSyntax = r.pileOfErrorMessages();
            if (NO_ERROR.CompareTo(patternSyntax) == 0)
            {
                ;
            }
            else
            {
                if (loud)
                {
                    Console.Error.WriteLine(patternSyntax);
                    System.Environment.Exit(ERROR);
                }
                else
                {
                    System.Environment.Exit(OTHERWISE);
                }

            }

            if (noListing)
            {
                if (standardInput)
                {
                    found = r.reading();
                    patternSyntax = r.pileOfErrorMessages();
                    if (NO_ERROR.CompareTo(patternSyntax) == 0)
                    {
                        if (found)
                        {
                            System.Environment.Exit(FOUND);
                        }
                        else
                        {
                            System.Environment.Exit(OTHERWISE);
                        }

                    }
                    else
                    {
                        if (loud)
                        {
                            Console.Error.WriteLine(patternSyntax);
                            System.Environment.Exit(ERROR);
                        }
                        else
                        {
                            System.Environment.Exit(OTHERWISE);
                        }

                    }

                }
                else
                {
                    try
                    {
                        FileInfo file = new FileInfo(FILE_NAME_LIST_TEXT);
                        long fileLength = file.Length;
                        if (fileLength < TOO_SMALL)
                        {
                            versionMessages(examples);
                            if (loud)
                            {
                                Console.Error.WriteLine("fileLength < TOO_SMALL");
                                System.Environment.Exit(ERROR);
                            }
                            else
                            {
                                System.Environment.Exit(OTHERWISE);
                            }

                        }

                    }
                    catch (ArgumentException e29)
                    {
                        Error.handler(e29.ToString());
                        versionMessages(examples);
                        if (loud)
                        {
                            Console.Error.WriteLine("ArgumentException e29");
                            System.Environment.Exit(ERROR);
                        }
                        else
                        {
                            System.Environment.Exit(OTHERWISE);
                        }

                    }

                }

            }
            else
            {
                if (newListShallBeProvided)
                {
                    provideFileNameListText(throughSubdirectories);
                }
                else
                {
                    versionMessages(examples);
                    if (loud)
                    {
                        System.Environment.Exit(ERROR);
                    }
                    else
                    {
                        System.Environment.Exit(OTHERWISE);
                    }

                }

                FileStream file;
                int napped = 0;
                long fileLength = 0;
                while (fileLength < TOO_SMALL)
                {
                    try
                    {
                        file = new FileStream(fileNameListText, FileMode.Open, FileAccess.Read);
                        fileLength = file.Length;
                        file.Close();
                    }
                    catch (IOException e30)
                    {
                        Error.handler(e30.ToString());
                    }

                    Thread.Sleep(WAITING_FOR_A_FILE);
                    napped += 1;
                    if (napped > manyTimes)
                    {
                        if (loud)
                        {
                            Console.Error.WriteLine("fileLength < TOO_SMALL");
                            System.Environment.Exit(ERROR);
                        }
                        else
                        {
                            System.Environment.Exit(OTHERWISE);
                        }

                    } // if

                } // while

            } // else

            if (specialFileName)
            {
                found = r.reading();
            }

            // Adapting the redirected text to Grep if -r and -u are specified (or if -g and -u are specified)
            if (throughSubdirectories)
            { // -r or -g
                if (windows)
                {
                    ;
                }
                else
                { // -u
                    UnixNotationAdapter u = new UnixNotationAdapter(fileNameListText, argument, loud, showRegularExpression);
                    if (reduceToCurrentDirectory) // -g
                    {
                        if (fullPath)
                        {
                            u.adaptingReducing();
                        }
                        else // -p-
                        {
                            u.closeSubdirectoryGate();
                        }

                    }
                    else
                    {
                        u.adapting();
                    }

                }

            }

            if (ticktock)
            {
                s.Stop();
                Console.Write(s.ElapsedTicks);
                Console.WriteLine(" ticks");
                w = Stopwatch.StartNew();
            }

            if (elapsed)
            {
                milliseconds = e.timeElapsed();
                if (noListing)
                {
                    if (ONE_MILLISECOND.CompareTo(milliseconds) == 0)
                    {
                        Console.WriteLine("One millisecond elapsed.");
                    }
                    else
                    {
                        Console.Write(milliseconds);
                        Console.WriteLine(" milliseconds elapsed.");
                    }

                }
                else
                {
                    if (ONE_MILLISECOND.CompareTo(milliseconds) == 0)
                    {
                        Console.Write("One millisecond elapsed before ");
                        Console.Write(fileNameListText);
                        Console.WriteLine(" was written.");
                    }
                    else
                    {
                        Console.Write(milliseconds);
                        Console.Write(" milliseconds elapsed before ");
                        Console.Write(fileNameListText);
                        Console.WriteLine(" was written.");
                    }

                }

            }

            switch (manager)
            {
                case BINARY_GREP:
                    exitCode = binaryGrep(r);
                    break;
                case ENCODED_GREP:
                    exitCode = encodedGrep(r);
                    break;
                case MIXED_GREP:
                    exitCode = mixedGrep(r);
                    break;
                default:
                    exitCode = asciiGrep(r);
                    break;
            }

            if (ticktock)
            {
                w.Stop();
                Console.Write( s.ElapsedTicks + w.ElapsedTicks );
                Console.WriteLine(" ticks");
            }

            if (elapsed)
            {
                milliseconds = e.timeElapsed();
                if (ONE_MILLISECOND.CompareTo(milliseconds) == 0)
                {
                    Console.WriteLine("One millisecond elapsed from start to finish.");
                }
                else
                {
                    Console.Write(milliseconds);
                    Console.WriteLine(" milliseconds elapsed from start to finish.");
                }

            }

            if (exitCode > 0)
            {
                System.Environment.Exit(exitCode);
            }

            patternSyntax = r.pileOfErrorMessages();
            if (NO_ERROR.CompareTo(patternSyntax) == 0)
            {
                if (found)
                {
                    exitCode = FOUND;
                }
                else
                {
                    exitCode = OTHERWISE;
                }

            }
            else
            {
                if (loud)
                {
                    Console.Error.WriteLine(patternSyntax);
                    exitCode = ERROR;

                }
                else
                {
                    if (found)
                    {
                        exitCode = FOUND;
                    }
                    else
                    {
                        exitCode = OTHERWISE;
                    }

                }

            }

            return exitCode;
        }

        private int useDIR()
        {
            int exitCode = SAFE_AND_SOUND;
            int read_size;

            try
            {
                Process p = new Process();
                p.StartInfo.FileName = CMD_EXE;
                
                if (throughSubdirectories)
                {
                    p.StartInfo.Arguments = C_DIR_S_B + wildCards;
                }
                else
                {
                    p.StartInfo.Arguments = C_DIR_B + wildCards;
                }

                p.StartInfo.UseShellExecute = false;
                p.StartInfo.RedirectStandardOutput = true;
                p.Start();
                StreamReader redirected = p.StandardOutput;
                StreamWriter listFile = new StreamWriter(fileNameListText);

                while (!redirected.EndOfStream)
                {
                    char_array = new char[CHAR_ARRAY_SIZE];
                    read_size = redirected.Read(char_array, 0, CHAR_ARRAY_SIZE);
                    listFile.Write(char_array, 0, read_size);
                }

                p.Close();

                listFile.Close();
            }
            catch (SecurityException e174)
            {
                exitCode = SECURITY_ERROR;
                Console.Error.WriteLine(e174.ToString());
            }
            catch (IOException e175)
            {
                exitCode = IO_ERROR;
                Console.Error.WriteLine(e175.ToString());
            }

            return exitCode;
        }

        private int use_ls()
        {
            int exitCode = SAFE_AND_SOUND;
            int read_size;

            try
            {
                Process p = new Process();
                p.StartInfo.FileName = CMD_EXE;

                if (throughSubdirectories)
                {
                    p.StartInfo.Arguments = C_LS_R_1;
                }
                else
                {
                    p.StartInfo.Arguments = C_LS_1 + wildCards;
                }

                p.StartInfo.UseShellExecute = false;
                p.StartInfo.RedirectStandardOutput = true;
                p.Start();
                StreamReader redirected = p.StandardOutput;
                StreamWriter listFile = new StreamWriter(fileNameListText);

                while (!redirected.EndOfStream)
                {
                    char_array = new char[CHAR_ARRAY_SIZE];
                    read_size = redirected.Read(char_array, 0, CHAR_ARRAY_SIZE);
                    listFile.Write(char_array, 0, read_size);
                }

                p.Close();

                listFile.Close();
            }
            catch (SecurityException e176)
            {
                exitCode = SECURITY_ERROR;
                Console.Error.WriteLine(e176.ToString());
            }
            catch (IOException e177)
            {
                exitCode = IO_ERROR;
                Console.Error.WriteLine(e177.ToString());
            }

            return exitCode;
        }

        private void addCodePage(string s, int q)
        {
            q += CP_SIZE;
            string tail = s.Substring(q);
            int code_page = sequence_to_integer(tail);
            d.Add(s, Encoding.GetEncoding(code_page));
        }

        private void versionMessages(bool examples)
        {
            StringBuilder commands = new StringBuilder();

            commands.Append("AceGrep Version "); commands.Append(VERSION); commands.Append(DOUBLE_SPACED);
            if (examples)
            {
                commands.Append("DIR FNL*.txt | AceGrep -c 'FNL'"); commands.Append(NEW_LINE);
                commands.Append("DIR FNL*.txt | AceGrep -c 'FNL' - bell.txt"); commands.Append(NEW_LINE);
                commands.Append("DIR FNL*.txt | AceGrep -c --label=DIR 'FNL' - bell.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep --ace \"a\" bell.txt"); commands.Append(NEW_LINE);

                // Shift_JIS
                commands.Append("AceGrep --ace -8 \"琴\" Samuel.txt > ACE"); commands.Append(NEW_LINE);
                commands.Append("AceGrep --ce -r \"fileNameListText\" \"*.cs\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep --encoding=CP1252 \"gem\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep --encoding=ISO-8859-1 \"gem\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep --help -q"); commands.Append(NEW_LINE);
                commands.Append("AceGrep --version"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -# -V"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -# \"a\" bell.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -8 \"ダビデ\" Samuel.txt | MORE"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -8 \"琴\" Samuel.txt > ACE"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -8 \"立琴\" Samuel.txt"); commands.Append(NEW_LINE);

                commands.Append("AceGrep -A2 \"barefoot\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -b- -c -f Q.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -B3 \"[Cc]lasp\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c 'no' sayno.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c --ascii \"99 00\" rannu.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c --elapse \"Loyalty\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c --in \"[Ff]ree\" \"[Pps]*.txt\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c --tick \"Loyalty\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -@ \"Loyalty\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -b- \"stone\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -b80 \"stone\" poem.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -e \"[Ss]tone\" \"[Ee]arring\" -e- poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -f patterns.txt poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -i -[ \"watcom\" atnd.exe"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -i \"[^s]wind[^o]\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -j \"[^dfp][Rr]ank\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -u -g -p- \"fame\" \"/^[^f].*\\.txt$/\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -u -r \"fame\" \"/^[^f].*\\.txt$/\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -u -r \"fame\" \"/^[^f].*\\.txt$/i\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -u -r \"[Ff]ree\" \"[Pps]*.txt\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -u -r --bat \"[Ff]ree\" \"[Pps]*.txt\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -u -r --re \"fame\" \"[!f]*.txt\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -v \"\\a\" i_43.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -v \"< 43\" i_43.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -x \"Hope.\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c -y \"99 00\" rannu.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c \"\\'no\\'\" sayno.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -c \"< 43\" i_43.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -C4 \"^Line-384\" LineNo.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -curj --re \"\\sgem\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -e \"[Ss]tone\" \"[Ee]arring\" -c poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -H -n -S \"rock of my refuge\" Psalms.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -H -n -T16 -S \"fortress\" \"p*.txt\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -H -n -T20 \"crown\" \"p*.txt\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -H -r \"Regex\" \"*.cs\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -L -q \"[Lloyalty\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -L -r \":.*P\" \"*.cs\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -l -r \"fileNameListText\" \"*.cs\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -m3 \"bird\" \"p*.txt\""); commands.Append(NEW_LINE);
                commands.Append("AceGrep -n -T6 \"[Rr]ock\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -n \"realm\" poems.txt"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -V"); commands.Append(NEW_LINE);
                commands.Append("AceGrep -v \"Psalms\" Psalms.txt");
            }
            else
            {
                commands.Append("AceGrep --help"); commands.Append(NEW_LINE);
            }

            string lines = commands.ToString();
            if (loud)
            {
                Console.Error.WriteLine(lines);
            }
            else
            {
                Console.WriteLine(lines);
            }

        }

        public string commandLine(string[] parameters)
        {
            bool regularExpression = true;
            bool versionExamples = false;
            int count = parameters.Length;
            int q = 0;
            Numbered n;
            StringBuilder or = new StringBuilder();

            foreach (string s in parameters)
            {
                if (string.IsNullOrEmpty(s))
                {
                    break;
                }

                switch (s)
                {
                    case ACE:
                        sideAce();
                        anotherRegularExpression = false;
                        break;
                    case ADDITIONAL_REGULAR_EXPRESSION_SEQUENCE_END:
                        anotherRegularExpression = false;
                        break;
                    case ADDITIONAL_REGULAR_EXPRESSION_SEQUENCE_START:
                        anotherRegularExpression = true;
                        break;
                    case ASCII_DASH:
                        manager = ASCII_GREP;
                        anotherRegularExpression = false;
                        break;
                    case ASCII_YES:
                        manager = ASCII_GREP;
                        anotherRegularExpression = false;
                        break;
                    case CODE_PAGE:
                        CHCP();
                        checkingCHCP_TEXT();
//                        anotherRegularExpression = false;
                        break;
                    case CONTROL_CHARACTERS:
                        manager = BINARY_GREP;
                        replaceControlCharacters = true;
                        anotherRegularExpression = false;
                        break;
                    case COUNT:
                        fileNamesRequired = true;
                        matchCount = true;
                        anotherRegularExpression = false;
                        break;
                    case COUNT_EQUALS_ZERO:
                        countEqualsZero = true;
                        fileNamesRequired = true;
                        matchCount = true;
                        anotherRegularExpression = false;
                        break;
                    case COUNT_EXCEEDS_ZERO:
                        countExceedsZero = true;
                        fileNamesRequired = true;
                        matchCount = true;
                        anotherRegularExpression = false;
                        break;
                    case CURRENT_DIRECTORY:
                        throughSubdirectories = true;
                        reduceToCurrentDirectory = true;
                        anotherRegularExpression = false;
                        break;
                    case DIR_TERM:
                        command = DIR;
                        dir = true;
                        windows = true;
                        anotherRegularExpression = false;
                        break;
                    case ELAPSE:
                        elapsed = true;
                        anotherRegularExpression = false;
                        break;
                    case EXTERNAL_BATCH:
                        externalBatch = true;
                        anotherRegularExpression = false;
                        break;
                    case FILE_NAMES_REQUIRED:
                        fileNamesRequired = true;
                        anotherRegularExpression = false;
                        break;
                    case FILE_NAME_FOR_EACH_MATCH:
                        fileNameForEachMatch = true;
                        anotherRegularExpression = false;
                        break;
                    case HELP:
                        versionExamples = true;
                        break;
                    case INSENSITIVE:
                        insensitive = true;
                        anotherRegularExpression = false;
                        break;
                    case INTERNAL_TERM:
                        command = INTERNAL;
                        dir = false;
                        windows = true;
                        anotherRegularExpression = false;
                        break;
                    case INVERTED:
                        inverted = true;
                        anotherRegularExpression = false;
                        break;
                    case JIKOKU:
                        TimeName ft = new TimeName(FILE_NAME_LIST, TEXT_FILE);
                        fileNameListText = ft.temporaryName(SERIAL_NUMBER);
                        anotherRegularExpression = false;
                        break;
                    case KANJI_8:
                        manager = ENCODED_GREP;
                        inputEncoding = UTF_8;
                        anotherRegularExpression = false;
                        break;
                    case KANJI_EUC_JP:
                        manager = ENCODED_GREP;
                        inputEncoding = EUC_JP;
                        anotherRegularExpression = false;
                        break;
                    case KANJI_SHIFT_JIS:
                        manager = ENCODED_GREP;
                        inputEncoding = SHIFT_JIS;
                        anotherRegularExpression = false;
                        break;
                    case KANJI_UTF_8:
                        manager = ENCODED_GREP;
                        inputEncoding = UTF_8;
                        anotherRegularExpression = false;
                        break;
                    case KANJI_WINDOWS31J:
                        manager = ENCODED_GREP;
                        inputEncoding = MS932;
                        anotherRegularExpression = false;
                        break;
                    case KUDA:
                        noListing = true;
                        standardInput = true;
                        anotherRegularExpression = false;
                        break;
                    case LINE_NUMBER:
                        lineNumberFollowed = true;
                        anotherRegularExpression = false;
                        break;
                    case MILLISECONDS:
                        elapsed = true;
                        anotherRegularExpression = false;
                        break;
                    case NO_BATCH:
                        noListing = true;
                        anotherRegularExpression = false;
                        break;
                    case NO_PATH:
                        fullPath = false;
                        anotherRegularExpression = false;
                        break;
                    case OBTAIN_PATTERNS_FROM_FILE:
                        listOfPatterns = true;
                        anotherRegularExpression = false;
                        break;
                    case QUIET:
                        loud = false;
                        anotherRegularExpression = false;
                        break;
                    case RECURSIVELY_LIST_ALL_FILES:
                        throughSubdirectories = true;
                        anotherRegularExpression = false;
                        break;
                    case SEPARATOR:
                        separator = obtainSeparatorFromFile(SEPARATE_TEXT);
                        anotherRegularExpression = false;
                        break;
                    case SHOW_REGULAR_EXPRESSION:
                        showRegularExpression = true;
                        anotherRegularExpression = false;
                        break;
                    case SPECIAL_FILE_NAME:
                        specialFileName = true;
                        anotherRegularExpression = false;
                        break;
                    case TEXT_EDITOR:
                        manager = BINARY_GREP;
                        textEditor = true;
                        anotherRegularExpression = false;
                        break;
                    case TICK:
                        ticktock = true;
                        anotherRegularExpression = false;
                        break;
                    case UNIX:
                        command = LS;
                        dir = false;
                        windows = false;
                        anotherRegularExpression = false;
                        break;
                    case WHOLE_LINE:
                        wholeLine = true;
                        anotherRegularExpression = false;
                        break;
                    case VERSION_PHRASE:
                        examples = false;
                        versionExamples = true;
                        break;
                    case VERSION_TERM:
                        examples = false;
                        versionExamples = true;
                        break;
                    default:
                        if (anotherRegularExpression)
                        {
                            or.Append(s);
                            or.Append(NEW_LINE);
                            regularExpressionCount += 1;
                            continue;
                        }

                        q = s.IndexOf(ENCODING_EQUALS);
                        if (q == 0)
                        {
                            manager = ENCODED_GREP;
                            q += ENCODING_EQUALS_SIZE;
                            inputEncoding = s.Substring(q);
                            continue;
                        }

                        q = s.IndexOf(LABEL_EQUALS);
                        if (q == 0)
                        {
                            q += LABEL_EQUALS_SIZE;
                            label = s.Substring(q);
                            continue;
                        }

                        q = s.IndexOf(BATCH);
                        if (q >= 0)
                        {
                            n = new Numbered(s);
                            manyTimes = n.number();
                            if (BATCH == n.parameter())
                            {
                                continue;
                            }

                            manyTimes = MANY_TIMES;
                        }

                        q = s.IndexOf(MAXIMUM);
                        if (q >= 0)
                        {
                            n = new Numbered(s);
                            maximumLineCount = n.number();
                            if (MAXIMUM == n.parameter())
                            {
                                continue;
                            }

                            maximumLineCount = MAXIMUM_LINE_COUNT;
                        }

                        q = s.IndexOf(TABLE);
                        if (q >= 0)
                        {
                            n = new Numbered(s);
                            fieldWidth = n.number();
                            if (TABLE == n.parameter())
                            {
                                continue;
                            }

                            fieldWidth = 0;
                        }

                        q = s.IndexOf(AFTER);
                        if (q >= 0)
                        {
                            n = new Numbered(s);
                            after = n.number();
                            if (AFTER.CompareTo(n.parameter()) == 0)
                            {
                                if (after == NOT_CONVERTED)
                                {
                                    after = 0;
                                }

                                ringCount = before + 1 + after;
                                if (RING_COUNT <= ringCount)
                                {
                                    ringCount = RING_COUNT;
                                }

                                continue;
                            }

                            after = 0;
                        }

                        q = s.IndexOf(BEFORE);
                        if (q >= 0)
                        {
                            n = new Numbered(s);
                            before = n.number();
                            if (BEFORE.CompareTo(n.parameter()) == 0)
                            {
                                if (before == NOT_CONVERTED)
                                {
                                    before = 0;
                                }

                                ringCount = before + 1 + after;
                                if (RING_COUNT <= ringCount)
                                {
                                    ringCount = RING_COUNT;
                                }

                                continue;
                            }

                            before = 0;
                        }

                        q = s.IndexOf(CONTEXT);
                        if (q >= 0)
                        {
                            n = new Numbered(s);
                            before = n.number();
                            if (CONTEXT.CompareTo(n.parameter()) == 0)
                            {
                                if (before == NOT_CONVERTED)
                                {
                                    before = 0;
                                }

                                after = before;
                                ringCount = before + 1 + after;
                                if (RING_COUNT <= ringCount)
                                {
                                    ringCount = RING_COUNT;
                                }

                                continue;
                            }

                            before = 0;
                        }

                        argument = s;
                        if (listOfPatterns)
                        {
                            if (regularExpression)
                            {
                                regularExpressionText = trimming(argument, NEITHER_CARET_NOR_DOLLAR);
                                regularExpression = false;
                                continue;
                            }

                        }
                        else
                        {
                            if (regularExpression)
                            {
                                re = trimming(argument, wholeLine);
                                regularExpression = false;
                                continue;
                            }

                        }

                        if (noListing)
                        {
                            continue;
                        }
                        else
                        {
                            deleteFileNameText();
                            newListShallBeProvided = true;
                            command = selected(dir, windows);
                            wildCards = argument;
                            if (command > 0 && externalBatch)
                            {
                                makeBatchFile(throughSubdirectories, argument, fileNameListText);
                            }

                        }

                        break;
                } // switch

            } // foreach

            Console.Write(NO_LOGO);
            if (versionExamples)
            {
                versionMessages(examples);
                if (loud)
                {
                    System.Environment.Exit(ERROR);
                }
                else
                {
                    System.Environment.Exit(OTHERWISE);
                }

            }

            if (noListing)
            {
                throughSubdirectories = false;
            }
            else
            {
                // If the last argument is not a file name, Grep uses the standard input.
                q = argument.IndexOf(QUOTE); // single quote
                if (q == 0)
                {
                    noListing = true;
                    standardInput = true;
                }

                q = argument.IndexOf(HYPHEN);
                if (q == 0)
                {
                    if (tryToOpenHyphenFile(argument))
                    {
                        deleteFileNameText();
                        newListShallBeProvided = true;
                        command = selected(dir, windows);
                        wildCards = argument;
                        if (command > 0 && externalBatch)
                        {
                            makeBatchFile(throughSubdirectories, argument, fileNameListText);
                        }

                    }
                    else
                    {
                        noListing = true;
                        standardInput = true;
                    }

                }

            }

            // PATTERNS_TEXT will be produced if the user types the following: -e "p[.][Bb]" "p[. ][Jj]" -e-
            if (regularExpressionCount > 0)
            {
                string lines = or.ToString();

                try
                {
                    StreamWriter patternList = new StreamWriter(PATTERNS_TEXT, NO_APPENDING, Encoding.Default);
                    patternList.Write(lines);
                    patternList.Close();
                }
                catch (IOException e31)
                {
                    if (loud)
                    {
                        Console.Error.WriteLine(e31.ToString());
                        System.Environment.Exit(ERROR);
                    }
                    else
                    {
                        System.Environment.Exit(OTHERWISE);
                    }

                }

                listOfPatterns = true;
                regularExpressionText = PATTERNS_TEXT;
                if (noListing)
                {
                    ;
                }
                else
                {
                    deleteFileNameText();
                    newListShallBeProvided = true;
                    command = selected(dir, windows);
                    wildCards = argument;
                    if (command > 0 && externalBatch)
                    {
                        makeBatchFile(throughSubdirectories, argument, fileNameListText);
                    }

                }

            }

            return argument;
        }

        private string obtainSeparatorFromFile(string fileName)
        {
            string s = COLON_GAP;

            try
            {
                StreamReader file = new StreamReader(fileName, Encoding.Default);
                s = file.ReadLine();
                if (s == null)
                {
                    s = COLON_GAP;
                }

                file.Close();
            }
            catch (IOException e32)
            {
                if (loud)
                {
                    Console.Error.WriteLine(e32.ToString());
                }

                preparingSeparateText();
            }

            return s;
        }

        public string[] singleCharacterParameters(string[] arguments)
        {
            bool done = false;
            int arraySize = arguments.Length;
            int compoundSize = 2;
            int k = 0;
            int parameterLength = 50 * arraySize;
            int q = 0;
            string c = EMPTY;
            string compound = EMPTY;
            string[] parameters = new string[parameterLength];

            for (int i = 0; i < arraySize; ++i)
            {
                if (k >= parameterLength)
                {
                    break;
                }

                if (string.IsNullOrEmpty(arguments[i]))
                {

                    break;
                }

                if (ADDITIONAL_REGULAR_EXPRESSION_SEQUENCE_END.CompareTo(arguments[i]) == 0)
                {
                    parameters[k++] = arguments[i];
                    continue;
                }

                if (ADDITIONAL_REGULAR_EXPRESSION_SEQUENCE_START.CompareTo(arguments[i]) == 0)
                {
                    parameters[k++] = arguments[i];
                    continue;
                }

                if (KANJI_EUC_JP.CompareTo(arguments[i]) == 0)
                {
                    parameters[k++] = arguments[i];
                    continue;
                }

                if (KANJI_SHIFT_JIS.CompareTo(arguments[i]) == 0)
                {
                    parameters[k++] = arguments[i];
                    continue;
                }

                if (KANJI_UTF_8.CompareTo(arguments[i]) == 0)
                {
                    parameters[k++] = arguments[i];
                    continue;
                }

                if (KANJI_WINDOWS31J.CompareTo(arguments[i]) == 0)
                {
                    parameters[k++] = arguments[i];
                    continue;
                }

                if (NO_BATCH.CompareTo(arguments[i]) == 0)
                {
                    parameters[k++] = arguments[i];
                    continue;
                }

                if (NO_PATH.CompareTo(arguments[i]) == 0)
                {
                    parameters[k++] = arguments[i];
                    continue;
                }

                if (SPECIAL_FILE_NAME.CompareTo(arguments[i]) == 0) // -
                {
                    parameters[k++] = arguments[i];
                    continue;
                }

                c = arguments[i].Substring(0, 1);
                if (HYPHEN.CompareTo(c) == 0)
                {
                    compound = arguments[i].Substring(1);
                    compoundSize = compound.Length;
                    if (compoundSize == 1) // a hyphen and a character
                    {
                        parameters[k++] = arguments[i];
                    }
                    else
                    {
                        c = arguments[i].Substring(1, 1);
                        if (HYPHEN.CompareTo(c) == 0) // two hyphens follewed by characters
                        {
                            parameters[k++] = arguments[i];
                        }
                        else
                        {
                            c = arguments[i].Substring(2, 1);
                            q = N_0TO9.IndexOf(c);
                            if (q >= 0) // -b40
                            {
                                parameters[k++] = arguments[i];
                            }
                            else
                            {
                                // Splitting a single-hyphen compound into many characters and adding a hyphen to each character
                                if (done)
                                {
                                    continue;
                                }
                                else
                                { // -curj
                                    for (int n = 0; n < compoundSize; ++n)
                                    {
                                        c = compound.Substring(n, 1);
                                        parameters[k++] = HYPHEN + c;
                                    }

                                    done = true;
                                }

                            }

                        }

                    }

                }
                else
                {
                    parameters[k++] = arguments[i];
                }

            }

            return parameters;
        }

        static private string trimming(string a, bool beginningEnd)
        {
            int size = a.Length;
            int size_1 = size - 1;
            int size_2 = size - 2;
            string head = a.Substring(0, 1);
            string tail = a.Substring(size_1, 1);
            string trimmed = a;

            if (QUOTE.CompareTo(tail) == 0)
            {
                if (QUOTE.CompareTo(head) == 0)
                {
                    trimmed = a.Substring(1, size_2);
                }
                else
                {
                    head = a.Substring(0, 2);
                    tail = a.Substring(size_2, 2);

                    if (YEN_QUOTE.CompareTo(tail) == 0)
                    {
                        if (YEN_QUOTE.CompareTo(head) == 0) // \'stone\'
                        {
                            trimmed = a.Substring(1, size_2 - 1) + QUOTE; // 'stone'
                        }

                    }

                }

            }

            if (beginningEnd)
            {
                size = trimmed.Length;
                size_1 = size - 1;
                head = trimmed.Substring(0, 1);
                tail = trimmed.Substring(size_1, 1);

                if (CARET.CompareTo(head) == 0)
                {
                    if (DOLLAR.CompareTo(tail) == 0)
                    {
                        ;
                    }
                    else
                    {
                        trimmed = trimmed + DOLLAR;
                    }

                }
                else
                {
                    if (DOLLAR.CompareTo(tail) == 0)
                    {
                        trimmed = CARET + trimmed;
                    }
                    else
                    {
                        trimmed = CARET_PARENTHESIS + trimmed + CLOSE_DOLLAR;
                    }

                }

            }

            return trimmed;
        }

    }
}
