﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace AceGrep
{
    public class MultiPatternFileNumberLineTable : Perusable
    {
        int column, nameLength, patternCount;
        Regex bellCompiled;
        Regex[] compiled;
        string wideSpace;
        string[] pattern;

        private int fixedWidthFontScreenColumn(string name)
        {
            int q;
            int size = 0;
            string c;

            for (int i = 0; i < name.Length; ++i)
            {
                c = name.Substring(i, 1);
                q = ASCII_KATAKANA.IndexOf(c);
                if (q >= 0)
                {
                    size += ASCII_SIZE;
                }
                else
                {
                    size += KANJI_SIZE;
                }

            }

            return size;
        }

        private string separating(string name, int lineNumber, string line)
        {
            string line_number = lineNumber.ToString();
            int line_number_length = line_number.Length;
            int q = column - line_number_length;
            int r = column - line_number_length - name.Length;
            int subtotal = nameLength + GAP_SIZE + line_number_length;
            string row;

            if (subtotal < column)
            {
                row = name + wideSpace.Substring(nameLength, r) + line_number + separator + line;
            }
            else
            {
                int abandoned;

                if (line_number_length < column)
                {
    // Grep abandons a part of the name so that the name-gap length can equal q.
                    string nameGap = name + GAP;
                    abandoned = nameGap.Length - q;
                    row = nameGap.Substring(abandoned) + line_number + separator + line;
                }
                else
                {
    // Grep abandons a part of the line number so that the line number length can equal column.
                    abandoned = line_number_length - column;
                    row = line_number.Substring(abandoned) + separator + line;
                }

            }

            return row;
        }

        public override bool reading()
        {
            bool b = false;
            string line = EMPTY;
            string name = label;
            nameLength = fixedWidthFontScreenColumn(name);

            try
            {
                for (int lineNumber = 1; ; lineNumber += 1)
                {
                    line = Console.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    foreach (Regex x in compiled)
                    {
                        m = x.Match(line);
                        if (inverted ^ m.Success)
                        {
                            b = true;
                            Console.WriteLine(separating(name, lineNumber, line));
                            break;
                        }

                    }

                }

            }
            catch (IOException e111)
            {
                Error.handler(e111.ToString());
                plus("IOException e111");
            }

            return b;
        }

        public override bool reading(string name)
        {
            bool b = false;
            string line = EMPTY;
            nameLength = fixedWidthFontScreenColumn(name);

            try
            {
                StreamReader file = new StreamReader(name, Encoding.Default);

                for (int lineNumber = 1; ; lineNumber += 1)
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    foreach (Regex x in compiled)
                    {
                        m = x.Match(line);
                        if (inverted ^ m.Success)
                        {
                            b = true;
                            Console.WriteLine(separating(name, lineNumber, line));
                            break;
                        }

                    }

                }

                file.Close();
            }
            catch (IOException e112)
            {
                Error.handler(e112.ToString());
                plus("IOException e112");
            }

            return b;
        }

        public override bool reading(string name, Encoding encoding)
        {
            bool b = false;
            string line = EMPTY;
            nameLength = fixedWidthFontScreenColumn(name);

            try
            {
                StreamReader file = new StreamReader(name, encoding);

                for (int lineNumber = 1; ; lineNumber += 1)
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    foreach (Regex x in compiled)
                    {
                        m = x.Match(line);
                        if (inverted ^ m.Success)
                        {
                            b = true;
                            Console.WriteLine(separating(name, lineNumber, line));
                            break;
                        }

                    }

                }

                file.Close();
            }
            catch (IOException e113)
            {
                Error.handler(e113.ToString());
                plus("IOException e113");
            }

            return b;
        }

        public MultiPatternFileNumberLineTable(string[] pn, int c, bool insensitive)
        {
            case_sensitive = !insensitive;

            compiling(BELL);
            bellCompiled = x; // x = new Regex(re)

            column = c;
            error = EMPTY;
            patternCount = pn.Length;
            wideSpace = EMPTY;

            compiled = new Regex[patternCount];
            pattern = new string[patternCount];
            for (int i = 0; i < column; ++i)
            {
                wideSpace += GAP;
            }

            try
            {
                Array.Copy(pn, 0, pattern, 0, patternCount);
            }
            catch (ArgumentException e114)
            {
                Error.handler(e114.ToString());
                plus("ArgumentException e114");
            }

            int error_contained = 0;
            for (int i = 0; i < patternCount && error_contained < MANY_TIMES; ++i)
            {
                if (compiling(pattern[i]))
                {
                    compiled[i] = x;
                }
                else
                {
                    compiled[i] = bellCompiled;
                    plus("/" + pattern[i] + "/");
                    error_contained += 1;
                }

            }

            if (error_contained == MANY_TIMES)
            {
                patternCount = MANY_TIMES;
            }

        }

    }
}
