﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace AceGrep
{
    public class MultiPatternNumberLineTable : Perusable
    {
        int column, patternCount;
        Regex bellCompiled;
        Regex[] compiled;
        string wideSpace;
        string[] pattern;

        private string separating(int lineNumber, string line)
        {
            string line_number = lineNumber.ToString();
            int line_number_length = line_number.Length;
            int q = column - line_number_length;
            string row;

            if (q > 0)
            { // line_number_length < column
                row = wideSpace.Substring(0, q) + line_number + separator + line;
            }
            else
            {
                // Grep abandons a part of the line number so that the line number length can equal column.
                int abandoned = -q; // line_number_length - column
                row = line_number.Substring(abandoned) + separator + line;
            }

            return row;
        }

        public override bool reading()
        {
            bool b = false;
            string line = EMPTY;

            try
            {
                for (int lineNumber = 1; ; lineNumber += 1)
                {
                    line = Console.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    foreach (Regex x in compiled)
                    {
                        m = x.Match(line);
                        if (inverted ^ m.Success)
                        {
                            b = true;
                            Console.WriteLine(separating(lineNumber, line));
                            break;
                        }

                    }

                }

            }
            catch (IOException e104)
            {
                Error.handler(e104.ToString());
                plus("IOException e104");
            }

            return b;
        }

        public override bool reading(string fileName)
        {
            bool b = false;

            try
            {
                StreamReader file = new StreamReader(fileName, Encoding.Default);
                string line = EMPTY;

                for (int lineNumber = 1; ; lineNumber += 1)
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    foreach (Regex x in compiled)
                    {
                        m = x.Match(line);
                        if (inverted ^ m.Success)
                        {
                            b = true;
                            Console.WriteLine(separating(lineNumber, line));
                            break;
                        }

                    }

                }

                file.Close();
            }
            catch (IOException e105)
            {
                Error.handler(e105.ToString());
                plus("IOException e105");
            }

            return b;
        }

        public override bool reading(string fileName, Encoding encoding)
        {
            bool b = false;

            try
            {
                StreamReader file = new StreamReader(fileName, encoding);
                string line = EMPTY;
                for (int lineNumber = 1; ; lineNumber += 1)
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    foreach (Regex x in compiled)
                    {
                        m = x.Match(line);
                        if (inverted ^ m.Success)
                        {
                            b = true;
                            Console.WriteLine(separating(lineNumber, line));
                            break;
                        }

                    }

                }

                file.Close();
            }
            catch (IOException e106)
            {
                Error.handler(e106.ToString());
                plus("IOException e106");
            }

            return b;
        }

        public MultiPatternNumberLineTable(string[] pn, int c, bool insensitive)
        {
            case_sensitive = !insensitive;

            compiling(BELL);
            bellCompiled = x; // x = new Regex(re)

            column = c;
            error = EMPTY;
            patternCount = pn.Length;
            wideSpace = EMPTY;

            compiled = new Regex[patternCount];
            pattern = new string[patternCount];
            for (int i = 0; i < column; ++i)
            {
                wideSpace += GAP;
            }

            try
            {
                Array.Copy(pn, 0, pattern, 0, patternCount);
            }
            catch (ArgumentException e107)
            {
                Error.handler(e107.ToString());
                plus("ArgumentException e107");
            }

            int error_contained = 0;
            for (int i = 0; i < patternCount && error_contained < MANY_TIMES; ++i)
            {
                if (compiling(pattern[i]))
                {
                    compiled[i] = x;
                }
                else
                {
                    compiled[i] = bellCompiled;
                    plus("/" + pattern[i] + "/");
                    error_contained += 1;
                }

            }

            if (error_contained == MANY_TIMES)
            {
                patternCount = MANY_TIMES;
            }

        }

    }
}
