﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace AceGrep
{
    public class NumberLineLimitedTable : Perusable
    {
        int column, maximumLineCount;
        string wideSpace;

        private string separating(int lineNumber, string line)
        {
            string line_number = lineNumber.ToString();
            int line_number_length = line_number.Length;
            int q = column - line_number_length;
            string row;

            if (q > 0)
            { // line_number_length < column
                row = wideSpace.Substring(0, q) + line_number + separator + line;
            }
            else
            {
// Grep abandons a part of the line number so that the line number length can equal column.
                int abandoned = -q; // line_number_length - column
                row = line_number.Substring(abandoned) + separator + line;
            }

            return row;
        }

        public override bool reading()
        {
            bool found = false;
            int count = 0;
            string line = EMPTY;

            try
            {
                for (int lineNumber = 1; count < maximumLineCount; lineNumber += 1)
                {
                    line = Console.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    m = x.Match(line);
                    if (inverted ^ m.Success)
                    {
                        Console.WriteLine( separating(lineNumber, line) );
                        count += 1;
                    }

                }

                if(count > 0) {
                    found = true;
                }

            }
            catch (IOException e124)
            {
                Error.handler(e124.ToString());
                plus("IOException e124");
            }

            return found;
        }

        public override bool reading(string fileName)
        {
            bool found = false;
            int count = 0;
            string line = EMPTY;

            try
            {
                StreamReader file = new StreamReader(fileName, Encoding.Default);

                for (int lineNumber = 1; count < maximumLineCount; lineNumber += 1)
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    m = x.Match(line);
                    if (inverted ^ m.Success)
                    {
                        Console.WriteLine( separating(lineNumber, line) );
                        count += 1;
                    }

                }

                file.Close();
                if(count > 0) {
                    found = true;
                }

            }
            catch (IOException e125)
            {
                Error.handler(e125.ToString());
                plus("IOException e125");
            }

            return found;
        }

        public override bool reading(string fileName, Encoding encoding)
        {
            bool found = false;
            int count = 0;
            string line = EMPTY;

            try
            {
                StreamReader file = new StreamReader(fileName, encoding);

                for (int lineNumber = 1; count < maximumLineCount; lineNumber += 1)
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    m = x.Match(line);
                    if (inverted ^ m.Success)
                    {
                        Console.WriteLine( separating(lineNumber, line) );
                        count += 1;
                    }

                }

                file.Close();
                if(count > 0) {
                    found = true;
                }

            }
            catch (IOException e126)
            {
                Error.handler(e126.ToString());
                plus("IOException e126");
            }

            return found;
        }

        public NumberLineLimitedTable(string re, int c, int m, bool insensitive)
        {
            case_sensitive = !insensitive;

            column = c;
            maximumLineCount = m;
            wideSpace = EMPTY;
            for (int i = 0; i < column; ++i)
            {
                wideSpace += GAP;
            }

            if (compiling(re))
            {
                error = EMPTY;
            }
            else
            {
                error = "compiling(re): /" + re + "/";
            }

        }

    }
}
