﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace AceGrep
{
    public class NumberLineTable : Perusable
    {
        int column;
        string wideSpace;

        private string separating(int lineNumber, string line)
        {
            string line_number = lineNumber.ToString();
            int line_number_length = line_number.Length;
            int q = column - line_number_length;
            string row;

            if (q > 0)
            { // line_number_length < column
                row = wideSpace.Substring(0, q) + line_number + separator + line;
            }
            else
            {
// Grep abandons a part of the line number so that the line number length can equal column.
                int abandoned = -q; // line_number_length - column
                row = line_number.Substring(abandoned) + separator + line;
            }

            return row;
        }

        public override bool reading()
        {
            bool b = false;
            string line = EMPTY;

            try
            {
                for (int lineNumber = 1; ; lineNumber += 1)
                {
                    line = Console.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    m = x.Match(line);
                    if (inverted ^ m.Success)
                    {
                        b = true;
                        Console.WriteLine( separating(lineNumber, line) );
                    }

                }

            }
            catch (IOException e101)
            {
                Error.handler(e101.ToString());
                plus("IOException e101");
            }

            return b;
        }

        public override bool reading(string fileName)
        {
            bool b = false;

            try
            {
                StreamReader file = new StreamReader(fileName, Encoding.Default);
                string line = EMPTY;

                for (int lineNumber = 1; ; lineNumber += 1)
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    m = x.Match(line);
                    if (inverted ^ m.Success)
                    {
                        b = true;
                        Console.WriteLine(separating(lineNumber, line));
                    }

                }

                file.Close();
            }
            catch (IOException e102)
            {
                Error.handler(e102.ToString());
                plus("IOException e102");
            }

            return b;
        }

        public override bool reading(string fileName, Encoding encoding)
        {
            bool b = false;

            try
            {
                StreamReader file = new StreamReader(fileName, encoding);
                string line = EMPTY;
                for (int lineNumber = 1; ; lineNumber += 1)
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    m = x.Match(line);
                    if (inverted ^ m.Success)
                    {
                        b = true;
                        Console.WriteLine(separating(lineNumber, line));
                    }

                }

                file.Close();
            }
            catch (IOException e103)
            {
                Error.handler(e103.ToString());
                plus("IOException e103");
            }

            return b;
        }

        public NumberLineTable(string re, int c, bool insensitive)
        {
            case_sensitive = !insensitive;

            column = c;
            wideSpace = EMPTY;
            for (int i = 0; i < column; ++i)
            {
                wideSpace += GAP;
            }

            if (compiling(re))
            {
                error = EMPTY;
            }
            else
            {
                error = "compiling(re): /" + re + "/";
            }

        }

    }
}
