﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace AceGrep
{
    public abstract class Perusable
    {
        protected const bool SERIAL_NUMBER = true;
        protected const byte CARET = 0x5E;
        protected const double THRESHOLD = 0.04; // 4%
        protected const int ASCII_SIZE = 1;
        protected const int BEGINNING = 2048;
        protected const int GAP_SIZE = 1;
        protected const int KANJI_SIZE = 2;
        protected const int MANY_TIMES = 40;
        protected const int NO_DATA = -1;
        protected const string ASCII = " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~";
        protected const string BELL = "\\a";
        protected const string CONTROL_CARET_CAPITALS = "CCC";
        protected const string DASH = "--";
        protected const string EMPTY = "";
        protected const string FROM_TO = " -> ";
        protected const string GAP = " ";
        protected const string I_O_EXCEPTION = "IOException";
        protected const string KATAKANA = "｡｢｣､･ｦｧｨｩｪｫｬｭｮｯｰｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐﾑﾒﾓﾔﾕﾖﾗﾘﾙﾚﾛﾜﾝﾞﾟ"; // UTF-8
        protected const string NO_ERROR = "no error";
        protected const string PUNCTUATION = " | ";
        protected const string STANDARD_INPUT = "stdin";
        protected const string TEXT_FILE = ".txt";

        protected bool case_sensitive, inverted;
        protected double rate;
        protected Match m;
        protected Regex x;
        protected string ASCII_KATAKANA = ASCII + KATAKANA;
        protected string error, label, re, separator;

        protected bool compiling(string re)
        {
            try
            {
                if (case_sensitive)
                {
                    x = new Regex(re);
                }
                else
                {
                    x = new Regex(re, RegexOptions.IgnoreCase);
                }

                return true;
            }
            catch (ArgumentException e45)
            {
                plus(e45.ToString());
                return false;
            }

        }

        public bool tooManyControlCharacters(string fileName)
        {
            bool b = false;

            try
            {
                byte[] buffer = File.ReadAllBytes(fileName);
                int count = 0;
                int other = 0;
                int size = buffer.Length;

                for (int i = 0; i < BEGINNING && i < size; ++i)
                {
                    switch (buffer[i])
                    {
                        case 0: count += 1; break;
                        case 1: count += 1; break;
                        case 2: count += 1; break;
                        case 3: count += 1; break;
                        case 4: count += 1; break;
                        case 5: count += 1; break;
                        case 6: count += 1; break;
                        case 7: other += 1; break;  // '\a' = 0x07
                        case 8: other += 1; break;  // '\b' = 0x08
                        case 9: other += 1; break;  // '\t' = 0x09
                        case 10: other += 1; break; // '\n' = 0x0A
                        case 11: other += 1; break; // '\v' = 0x0B
                        case 12: other += 1; break; // '\f' = 0x0C
                        case 13: other += 1; break; // '\r' = 0x0D
                        case 14: count += 1; break;
                        case 15: count += 1; break;
                        case 16: count += 1; break;
                        case 17: count += 1; break;
                        case 18: count += 1; break;
                        case 19: count += 1; break;
                        case 20: count += 1; break;
                        case 21: count += 1; break;
                        case 22: count += 1; break;
                        case 23: count += 1; break;
                        case 24: count += 1; break;
                        case 25: count += 1; break;
                        case 26: count += 1; break;
                        case 27: count += 1; break;
                        case 28: count += 1; break;
                        case 29: count += 1; break;
                        case 30: count += 1; break;
                        case 31: count += 1; break;
                        default:
                            other += 1;
                            break;
                    }

                }

                double control_characters = (double)count;
                double other_characters = (double)other;
                double sum = (double)(count + other);
                rate = control_characters / sum;
                if (rate > THRESHOLD)
                {
                    b = true;
                }

            }
            catch (IOException e46)
            {
                plus("IOException e46");
                Error.handler(e46.ToString());
            }

            return b;
        }

        public double controlCharacterRate()
        {
            return rate;
        }

        public string pileOfErrorMessages()
        {
            if (error.Length == 0)
            {
                error = NO_ERROR;
            }

            return error;
        }

        public string plus(string again)
        {
            error += PUNCTUATION;
            error += again;
            return error;
        }

        public string replaceControlCharactersWithCaretCapitals(string fileName)
        {
            try
            {
                byte[] buffer = File.ReadAllBytes(fileName);
                int k = 0;
                int size = buffer.Length;
                byte[] largeBuffer = new byte[2 * size + 1];

                foreach (byte c in buffer)
                {
                    switch (c)
                    {
                        case 0: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'@'; break;
                        case 1: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'A'; break;
                        case 2: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'B'; break;
                        case 3: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'C'; break;
                        case 4: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'D'; break;
                        case 5: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'E'; break;
                        case 6: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'F'; break;
                        case 7: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'G'; break; // '\a' = 0x07
                        case 8: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'H'; break; // '\b' = 0x08
                        case 11: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'K'; break; // '\v' = 0x0B
                        case 12: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'L'; break; // '\f' = 0x0C
                        case 14: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'N'; break;
                        case 15: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'O'; break;
                        case 16: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'P'; break;
                        case 17: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'Q'; break;
                        case 18: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'R'; break;
                        case 19: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'S'; break;
                        case 20: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'T'; break;
                        case 21: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'U'; break;
                        case 22: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'V'; break;
                        case 23: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'W'; break;
                        case 24: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'X'; break;
                        case 25: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'Y'; break;
                        case 26: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'Z'; break;
                        case 27: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'['; break;
                        case 28: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'\\'; break;
                        case 29: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)']'; break;
                        case 30: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'^'; break;
                        case 31: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'_'; break;
                        case 127: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'7'; break; // 0x7F = 127
                        case 253: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'3'; break; // 0xFD = -3
                        case 254: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'2'; break; // 0xFE = -2
                        case 255: largeBuffer[k++] = CARET; largeBuffer[k++] = (byte)'1'; break; // 0xFF = -1
                        default:
                            largeBuffer[k++] = c;
                            break;
                    }

                }

                largeBuffer[k] = 0;

                TimeName ct = new TimeName(CONTROL_CARET_CAPITALS, TEXT_FILE);
                string caretCapitalsText = ct.temporaryName(SERIAL_NUMBER);
                File.WriteAllBytes(caretCapitalsText, largeBuffer);

                Console.WriteLine(fileName + FROM_TO + caretCapitalsText);
                return caretCapitalsText;
            }
            catch (IOException e47)
            {
                Error.handler(e47.ToString());
                plus("IOException e47");
                return I_O_EXCEPTION;
            }

        }

        public void setting(string l, string s, bool i)
        {
            label = l;
            separator = s;
            inverted = i;
        }

        public abstract bool reading();
        public abstract bool reading(string fileName);
        public abstract bool reading(string fileName, Encoding encoding);
    }
}
