﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace AceGrep
{
    class UnixNotationAdapter
    {
        const bool NO_APPENDING = false;
        const int CURRENT_DIRECTORY_SIZE = 1; // "."
        const int SLASH_SIZE = 1;
        const string CURRENT_DIRECTORY = ".";
        const string DIRECTORY_NAME_END = ":";
        const string FILE_NAME_LIST_TEXT = "fileName.txt";
        const string NEW_LINE = "\x0D\x0A";
        const string REGULAR_EXPRESSION_COLON = "Regular expression: ";
        const string SEPARATOR = "\\";
        const string SLASH = "/";

        bool loud = true;
        bool showRegularExpression;
        Regex compiled; // fileExpressedAsRegularExpression()
        string fileNameListText = FILE_NAME_LIST_TEXT;
        string argument, USER_DIRECTORY;

        private bool fileExpressedAsRegularExpression(string re, bool case_sensitive)
        {
            bool b = true;

            try
            {
                if (case_sensitive)
                {
                    compiled = new Regex(re);
                }
                else
                {
                    compiled = new Regex(re, RegexOptions.IgnoreCase);
                }

            }
            catch (ArgumentException e53)
            {
                Error.handler(e53.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("ArgumentException e53");
                }

                b = false;
            }

            return b;
        }

        public bool adapting()
        {
            bool b = false;
            int q;
            Match m;
            string c, body, line;
            string directory = USER_DIRECTORY;
            StringBuilder p;
            StringBuilder pathName = new StringBuilder();

            WildCardToRegularExpression w = new WildCardToRegularExpression(argument);
            if (showRegularExpression)
            {
                Console.Write(REGULAR_EXPRESSION_COLON);
                Console.WriteLine(w.converted());
            }

            if (fileExpressedAsRegularExpression(w.converted(), w.caseSensitive()))
            {
                ;
            }
            else
            {
                if (loud)
                {
                    Console.Error.WriteLine("adapting(): fileExpressedAsRegularExpression( w.converted(), w.caseSensitive()");
                }

                return b;
            }

            try
            {
                StreamReader file = new StreamReader(fileNameListText, Encoding.Default);
                int current_directory_slash_size;
                for (; ; )
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    q = line.IndexOf(CURRENT_DIRECTORY);
                    if (q == 0)
                    {
                        q = line.IndexOf(DIRECTORY_NAME_END);
                        if (q > CURRENT_DIRECTORY_SIZE)
                        {
                            p = new StringBuilder();

                            current_directory_slash_size = CURRENT_DIRECTORY_SIZE + SLASH_SIZE;
                            body = line.Substring(current_directory_slash_size, q - current_directory_slash_size);
                            for (int i = 0; i < body.Length; ++i)
                            {
                                c = body.Substring(i, 1);
                                if (SLASH.CompareTo(c) == 0)
                                {
                                    p.Append(SEPARATOR);
                                }
                                else
                                {
                                    p.Append(c);
                                }

                            }

                            directory = USER_DIRECTORY + p.ToString() + SEPARATOR;
                        }
                        else
                        {
                            directory = USER_DIRECTORY;
                        }

                        continue;
                    }

                    m = compiled.Match(line);
                    if (m.Success)
                    {
                        pathName.Append(directory);
                        pathName.Append(line);
                        pathName.Append(NEW_LINE);
                    }

                }

                file.Close();

                StreamWriter text = new StreamWriter(fileNameListText);
                string listOfFileNames = pathName.ToString();
                text.Write(listOfFileNames);
                text.Close();
                b = true;
            }
            catch (IOException e54)
            {
                Error.handler(e54.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e54");
                }

            }

            return b;
        }

        public bool adaptingReducing()
        {
            bool b = false;
            int q;
            Match m;
            string line;
            StringBuilder pathName = new StringBuilder();

            WildCardToRegularExpression w = new WildCardToRegularExpression(argument);
            if (w.throughSubdirectories())
            {
                return adapting();
            }

            if (showRegularExpression)
            {
                Console.Write(REGULAR_EXPRESSION_COLON);
                Console.WriteLine(w.converted());
            }

            if (fileExpressedAsRegularExpression(w.converted(), w.caseSensitive()))
            {
                ;
            }
            else
            {
                if (loud)
                {
                    Console.Error.WriteLine("adaptingReducing(): fileExpressedAsRegularExpression( w.converted(), w.caseSensitive()");
                }

                return b;
            }

            try
            {
                StreamReader file = new StreamReader(fileNameListText, Encoding.Default);
                for (; ; )
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    q = line.IndexOf(CURRENT_DIRECTORY);
                    if (q == 0)
                    {
                        q = line.IndexOf(DIRECTORY_NAME_END);
                        if (q > CURRENT_DIRECTORY_SIZE)
                        {
                            break;
                        }

                        continue;
                    }

                    m = compiled.Match(line);
                    if (m.Success)
                    {
                        pathName.Append(USER_DIRECTORY);
                        pathName.Append(line);
                        pathName.Append(NEW_LINE);
                    }

                }

                file.Close();

                StreamWriter text = new StreamWriter(fileNameListText, NO_APPENDING, Encoding.Default);
                string listOfFileNames = pathName.ToString();
                text.Write(listOfFileNames);
                text.Close();
                b = true;
            }
            catch (IOException e55)
            {
                Error.handler(e55.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e55");
                }

            }

            return b;
        }

        public bool closeSubdirectoryGate()
        {
            bool b = false;
            int q;
            Match m;
            string line;
            StringBuilder fileName = new StringBuilder();

            WildCardToRegularExpression w = new WildCardToRegularExpression(argument);
            if (w.throughSubdirectories())
            {
                return adapting();
            }

            if (showRegularExpression)
            {
                Console.Write(REGULAR_EXPRESSION_COLON);
                Console.WriteLine(w.converted());
            }

            if (fileExpressedAsRegularExpression(w.converted(), w.caseSensitive()))
            {
                ;
            }
            else
            {
                if (loud)
                {
                    Console.Error.WriteLine("adaptingReducing(): fileExpressedAsRegularExpression( w.converted(), w.caseSensitive()");
                }

                return b;
            }

            try
            {
                StreamReader file = new StreamReader(fileNameListText, Encoding.Default);
                for (; ; )
                {
                    line = file.ReadLine();
                    if (line == null)
                    {
                        break;
                    }

                    q = line.IndexOf(CURRENT_DIRECTORY);
                    if (q == 0)
                    {
                        q = line.IndexOf(DIRECTORY_NAME_END);
                        if (q > CURRENT_DIRECTORY_SIZE)
                        {
                            break;
                        }

                        continue;
                    }

                    m = compiled.Match(line);
                    if (m.Success)
                    {
                        fileName.Append(line);
                        fileName.Append(NEW_LINE);
                    }

                }

                file.Close();

                StreamWriter text = new StreamWriter(fileNameListText, NO_APPENDING, Encoding.Default);
                string listOfFileNames = fileName.ToString();
                text.Write(listOfFileNames);
                text.Close();
                b = true;
            }
            catch (IOException e56)
            {
                Error.handler(e56.ToString());
                if (loud)
                {
                    Console.Error.WriteLine("IOException e56");
                }

            }

            return b;
        }

        public UnixNotationAdapter(string f, string a, bool l, bool s)
        {
            argument = a;
            fileNameListText = f;
            loud = l;
            showRegularExpression = s;
            USER_DIRECTORY = Directory.GetCurrentDirectory() + SEPARATOR;
        }

    }
}
