﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace AceGrep
{
    class WildCardToRegularExpression
    {
        const int DOT_SIZE = 1;
        const int INSENSITIVE_SIZE = 2;
        const int SLASH_SIZE = 1;
        const int STAR_SIZE = 1;
        const string BRACKET = "[";
        const string BRACKET_CARET = "[^";
        const string BRACKET_EXCLAMATION = "[!";
        const string BACKSLASH_CARET = "\\^";
        const string BACKSLASH_DOLLAR = "\\$";
        const string BACKSLASH_DOT = "\\.";
        const string CARET = "^";
        const string DOLLAR = "$";
        const string DOT = ".";
        const string DOT_STAR = ".*";
        const string EMPTY = "";
        const string INSENSITIVE = "/i";
        const string QUESTION = "?";
        const string SLASH = "/";
        const string STAR = "*";
        bool case_sensitive, enclosed, included, subdirectories;
        string re = EMPTY;
        string sequence_of_dots = EMPTY;

        private bool headQuestionsDotTail(string sequence)
        {
            bool b = false;
            string body, head, save, tail;
            int q;

            q = sequence.IndexOf(QUESTION);
            if (q >= 0)
            {
                head = sequence.Substring(0, q);
                tail = sequence.Substring(q);
                q = tail.LastIndexOf(QUESTION);
                if (q >= 0)
                {
                    body = tail.Substring(0, q);
                    tail = tail.Substring(q);

                    if (metaCharacterIncluded(head))
                    {
                        included = true;
                    }
                    else
                    {
                        if (metaCharacterIncluded(tail))
                        {
                            included = true;
                            save = head;
                            head = tail.Substring(0, DOT_SIZE);
                            tail = tail.Substring(DOT_SIZE);

                            if (DOT.CompareTo(head) == 0)
                            {
                                if (metaCharacterIncluded(tail))
                                {
                                    included = true;
                                }
                                else
                                {
                                    re = CARET + save + BACKSLASH_DOT + tail + DOLLAR;
                                    b = true;
                                }

                            }

                        }
                        else
                        {
                            if (sequenceOfQuestionMarks(body))
                            {
                                re = CARET + head + sequence_of_dots + tail + DOLLAR;
                                b = true;
                            }

                        }

                    }

                }

            }

            return b;
        }

        private bool headStarDotTail(string sequence)
        {
            bool b = false;
            string head, save, tail;
            int q;

            q = sequence.IndexOf(STAR);
            if (q >= 0)
            {
                head = sequence.Substring(0, q);
                q += STAR_SIZE;
                tail = sequence.Substring(q);

                if (metaCharacterIncluded(head))
                {
                    included = true;
                }
                else
                {
                    if (metaCharacterIncluded(tail))
                    {
                        included = true;
                        save = head;
                        head = tail.Substring(0, DOT_SIZE);
                        tail = tail.Substring(DOT_SIZE);

                        if (DOT.CompareTo(head) == 0)
                        {
                            if (metaCharacterIncluded(tail))
                            {
                                included = true;
                            }
                            else
                            {
                                re = CARET + save + DOT_STAR + BACKSLASH_DOT + tail + DOLLAR;
                                b = true;
                            }

                        }

                    }
                    else
                    {
                        re = CARET + head + DOT_STAR + tail + DOLLAR;
                        b = true;
                    }

                }

            }

            return b;
        }

        private bool metaCharacterIncluded(string sequence)
        {
            bool included = false;
            string[] meta = { "\\", "^", "$", ".", "[", "]", "*", "+", "?" };
            string c;

            for (int m = 0; m < meta.Length; ++m)
            {
                for (int i = 0; i < sequence.Length; ++i)
                {
                    c = sequence.Substring(i, 1);
                    if (c.CompareTo(meta[m]) == 0)
                    {
                        included = true;
                        break;
                    }

                }

            }

            return included;
        }

        private bool regularExpressionEnclosed(string sequence)
        {
            case_sensitive = true;
            enclosed = false;
            int sequenceLength = sequence.Length;
            if (sequenceLength < 2)
            {
                return enclosed;
            }

            string c = sequence.Substring(0, SLASH_SIZE);
            if (SLASH.CompareTo(c) == 0)
            {
                int q = sequenceLength - SLASH_SIZE;
                c = sequence.Substring(q);
                if (SLASH.CompareTo(c) == 0)
                {
                    case_sensitive = true;
                    enclosed = true;
                    re = sequence.Substring(SLASH_SIZE, q - SLASH_SIZE);
                }
                else
                {
                    q = sequenceLength - INSENSITIVE_SIZE;
                    string tail = sequence.Substring(q);
                    if (INSENSITIVE.CompareTo(tail) == 0)
                    {
                        case_sensitive = false;
                        enclosed = true;
                        re = sequence.Substring(SLASH_SIZE, q - SLASH_SIZE);
                    }

                }

            }

            return enclosed;
        }

        private bool sequenceOfQuestionMarks(string sequence)
        {
            bool intermittent = false;
            string c;
            StringBuilder qm = new StringBuilder();

            for (int i = 0; i < sequence.Length; ++i)
            {
                c = sequence.Substring(i, 1);
                if (QUESTION.CompareTo(c) == 0)
                {
                    qm.Append(DOT);
                    continue;
                }
                else
                {
                    intermittent = true;
                    break;
                }

            }

            sequence_of_dots = qm.ToString();
            return !intermittent;
        }

        private bool starBodyStar(string sequence)
        {
            bool b = false;
            string body;

            int sequenceLength = sequence.Length;
            if (sequenceLength < 2)
            {
                return b;
            }

            string c = sequence.Substring(0, STAR_SIZE);
            if (STAR.CompareTo(c) == 0)
            {
                int q = sequenceLength - STAR_SIZE;
                c = sequence.Substring(q);
                if (STAR.CompareTo(c) == 0)
                {
                    body = sequence.Substring(STAR_SIZE, q - STAR_SIZE);
                    if (metaCharacterIncluded(body))
                    {
                        included = true;
                        if (DOT.CompareTo(body) == 0)
                        {
                            re = CARET + DOT_STAR + BACKSLASH_DOT + DOT_STAR + DOLLAR;
                            b = true;
                        }

                    }
                    else
                    {
                        re = CARET + DOT_STAR + body + DOT_STAR + DOLLAR;
                        b = true;
                    }

                }

            }

            return b;
        }

        private bool typical(string sequence)
        {
            if (headStarDotTail(sequence))
            {
                return true;
            }

            if (starBodyStar(sequence))
            {
                return true;
            }

            if (headQuestionsDotTail(sequence))
            {
                return true;
            }

            if (DOT.CompareTo(sequence) == 0)
            {
                re = CARET + DOT_STAR + DOLLAR;
                subdirectories = true;
                return true;
            }

            if (STAR.CompareTo(sequence) == 0)
            {
                re = CARET + DOT_STAR + DOLLAR;
                return true;
            }

            return false;
        }

        private void replaceWildCardsOneByOne(string sequence)
        {
            string c, negation;
            StringBuilder obo = new StringBuilder(); // one by one
            obo.Append(CARET);

            for (int i = 0; i < sequence.Length; ++i)
            {
                c = sequence.Substring(i, 1);
                if (BRACKET.CompareTo(c) == 0)
                {
                    negation = sequence.Substring(i, 2);
                    if (BRACKET_EXCLAMATION.CompareTo(negation) == 0)
                    {
                        obo.Append(BRACKET_CARET);
                        ++i;
                    }
                    else
                    {
                        obo.Append(BRACKET);
                    }

                }
                else if (CARET.CompareTo(c) == 0)
                {
                    obo.Append(BACKSLASH_CARET);
                }
                else if (DOLLAR.CompareTo(c) == 0)
                {
                    obo.Append(BACKSLASH_DOLLAR);
                }
                else if (DOT.CompareTo(c) == 0)
                {
                    obo.Append(BACKSLASH_DOT);
                }
                else if (QUESTION.CompareTo(c) == 0)
                {
                    obo.Append(DOT);
                }
                else if (STAR.CompareTo(c) == 0)
                {
                    obo.Append(DOT_STAR);
                }
                else
                {
                    obo.Append(c);
                }

            }

            obo.Append(DOLLAR);
            re = obo.ToString();
        }

        public bool metaCharacter()
        {
            return included;
        }

        public bool regularExpression()
        {
            return enclosed;
        }

        public bool throughSubdirectories()
        {
            return subdirectories;
        }

        public bool caseSensitive()
        {
            return case_sensitive;
        }

        public string converted()
        {
            return re;
        }

        public WildCardToRegularExpression(string argument)
        {
            case_sensitive = true;

            if (metaCharacterIncluded(argument))
            {
                included = true;
                if (regularExpressionEnclosed(argument))
                {
                    ;
                }
                else
                {
                    if (typical(argument))
                    {
                        ;
                    }
                    else
                    {
                        replaceWildCardsOneByOne(argument);
                    }

                }

            }
            else
            {
                re = CARET + argument + DOLLAR;
            }

        }

    }
}
