#include "atnd.h"

/*
atnd 1.03
Showing lines at and around the specified point (AT and arouND)
Copyright 2015 Sogaya

WATCOM

*/

extern char *queue[QUEUE_ARRAY_SIZE];
extern int queue_begin;
static const int RING_COUNT = QUEUE_ARRAY_SIZE - 1;

char *parameter_of_numbered(char *a, int *number) {
    char *stop;
    char numerical_sequence[INTEGER_SIZE];
    int non_number_length, span;
    int length = strlen(a);
    char *p = malloc( length + 1 );

/* Copying all */
    strcpy(p, a);
    span = strspn(strrev(p), "0123456789");
    non_number_length = length - span;

/* Copying the second part of a */
    memcpy(numerical_sequence, a + non_number_length, span);
    numerical_sequence[span] = '\0';
    *number = (int)strtol(numerical_sequence, &stop, 10);

/* Copying the first part of a */
    memcpy(p, a, non_number_length);
    p[non_number_length] = '\0';
    return p;
}

int line_feed_check(char *file_name, LF_CHECK *check) {
    FILE *file;
    int c = '\0';
    int span = 0;

    check->count = 0;
    check->max = 0;
    file = fopen(file_name, "rt");
    if(file == NULL) {
        fprintf(stderr, "fopen(file_name, \"rt\"): %s\n", file_name);
        exit(1);
    }

    while(c != EOF) {
        c = fgetc(file);
        switch(c) {
        case EOF:
            break;
        case '\n': /* LF */
/* Counting lines */
            check->count += 1;

/* Recording the maximum span from LF to LF */
            span += 1;
            if(span > check->max) {
                check->max = span;
                check->line_number = check->count;
            }

            span = 0;
            break;
        default:
/* Counting bytes */
            span += 1;
            break;
        }

    }

    fclose(file);
    return check->count;
}

int main(int argc, char *argv[]) {
    BOOL count_lines = 0; /* FALSE */
    BOOL written = 0; /* FALSE */
    char *buffer;
    char *non_number;
    char *p;
    char file_name[FILE_NAME_SIZE];
    const BOOL WITHOUT_COMMENTS = 0; /* FALSE */
    const BOOL WITH_COMMENTS = TRUE;
    const int PARAMETER_SIZE = 2;
    const char AFTER[] = "-a";
    const char BEFORE[] = "-b";
    const char LARGE_SIZE[] = "-s";
    const char LINE_FEED[] = "-LF";
    const char CONTEXT[] = "-c";
    const char POINT[] = "-p";
    const char version[] = "-V";
    const char version_comment[] = "-VC";
    FILE *file;
    int after = 0;
    int before = 0;
    int buffer_size = BUFFER_SIZE;
    int end_of_context, i, line, number, specified;
    int ring_count = 1;
    LF_CHECK line_feed;
    LF_CHECK *check = &line_feed;

/* command line */
    for(i = 1; i < argc; i++) {
        if(! strcmp(LINE_FEED, argv[i]) ) {
            count_lines = TRUE;
        } else if(! strcmp(version, argv[i]) ) {
            versionMessage(WITHOUT_COMMENTS);
            return 1;

        } else if(! strcmp(version_comment, argv[i]) ) {
            versionMessage(WITH_COMMENTS);
            return 1;

        } else {
            if( strlen(argv[i]) > PARAMETER_SIZE ) {
                non_number = parameter_of_numbered(argv[i], &number);

                if(! strcmp(AFTER, non_number) ) {
                    after = number;
                    ring_count = before + 1 + after;
                    if(RING_COUNT <= ring_count) {
                        ring_count = RING_COUNT;
                    }

                    continue;
                } else if(! strcmp(BEFORE, non_number) ) {
                    before = number;
                    ring_count = before + 1 + after;
                    if(RING_COUNT <= ring_count) {
                        ring_count = RING_COUNT;
                    }

                    continue;
                } else if(! strcmp(CONTEXT, non_number) ) {
                    after = number;
                    before = after;
                    ring_count = before + 1 + after;
                    if(RING_COUNT <= ring_count) {
                        ring_count = RING_COUNT;
                    }

                    continue;
                } else if(! strcmp(LARGE_SIZE, non_number) ) {
                    buffer_size = number;
                    if(buffer_size < BUFFER_SIZE) {
                        buffer_size = BUFFER_SIZE;
                    }

                    continue;
                } else if(! strcmp(POINT, non_number) ) {
                    specified = number;
                    continue;
                }

                free(non_number);
            }

            strcpy(file_name, argv[i]);
        }

    }

    if(count_lines) {
        line_feed_check(file_name, check);
        fprintf(stdout, "%s: %d lines\n", file_name, check->count);
        buffer_size = check->max;
        fprintf(stdout, "maximum LF-to-LF span of %d bytes (line-%d)\n", buffer_size, check->line_number);
        buffer_size += 1;
    }

/* fgets buffer */
    buffer = malloc(buffer_size);
    if(buffer == NULL) {
        fprintf(stderr, "main(): buffer_size = %d\n", buffer_size);
        return 1;
    }

/* ring buffer */
    ring_count = produce_ring_buffer(ring_count, buffer_size);

    file = fopen(file_name, "rt");
    if(file == NULL) {
        fprintf(stderr, "%s\n", file_name);
        return 1;
    }

    end_of_context = specified + after;
    for(line = 1; ; line++) {
        p = fgets(buffer, buffer_size, file);
        if(p == NULL) {
            break;
        }

        chopped(buffer);
        pop_front();
        push_back(buffer, buffer_size);
        if(line == end_of_context) {
            showRing(ring_count, buffer_size);
            written = TRUE;
        }

    }

    fclose(file);
    free(buffer);
    clear();

    if(written) {
        return 0;
    } else {
        showRing(ring_count, buffer_size);
    }

    return 1;
}

void chopped(char *s) {
    char *p;

/* Removing the new-line character */
    p = strrchr(s, '\n');
    if(p != NULL) {
        *p = '\0';
    }
}

void showRing(int repetition, int buffer_size) {
    char *buffer;
    char *q;
    const char NEW_LINE[] = "\n";
    int item = queue_begin;
    int k;

    if( repetition > queue_size() ) {
        repetition = queue_size();
    }

    buffer = malloc( buffer_size * repetition );
    if(buffer == NULL) {
        fprintf(stderr, "showRing(): buffer_size = %d ( repetition = %d )\n", buffer_size, repetition);
        return;
    }

    buffer[0] = '\0';
    for(k = 0; k < repetition; k++) {
        q = queue[item];
        strcat(buffer, q);
        strcat(buffer, NEW_LINE);
        item = next(item);
    }

    fputs(buffer, stdout);
    free(buffer);
}

void versionMessage(BOOL comments) {

    if(comments) {
        fprintf(stderr, "atnd %1.2f\n\n", VERSION);
        fprintf(stderr, "atnd -a2 -p567 file_name # 2 lines after line-567 \n");
        fprintf(stderr, "atnd -b3 -p567 file_name # 3 lines before line-567 \n");
        fprintf(stderr, "atnd -c4 -p567 file_name # 4 lines before line-567 and 4 lines after line-567 \n");
        fprintf(stderr, "atnd -LF -p567 file_name # Count lines. Record the maximum span from LF to LF. \n");
        fprintf(stderr, "atnd -p567 file_name # line-567 \n");
        fprintf(stderr, "atnd -s10240 -p567 file_name # large buffer size of 10240 bytes \n");
        fprintf(stderr, "atnd -V # version \n");
        fprintf(stderr, "atnd -VC # version with comments \n");
    } else {
        fprintf(stderr, "atnd %1.2f\n\n", VERSION);
        fprintf(stderr, "atnd -a2 -p567 file_name \n");
        fprintf(stderr, "atnd -b3 -p567 file_name \n");
        fprintf(stderr, "atnd -c4 -p567 file_name \n");
        fprintf(stderr, "atnd -LF -p567 file_name \n");
        fprintf(stderr, "atnd -p567 file_name \n");
        fprintf(stderr, "atnd -s10240 -p567 file_name \n");
        fprintf(stderr, "atnd -V \n");
        fprintf(stderr, "atnd -VC \n");
    }

}

